/*************************************************************************
                      ganalyse.h

equivalence classes, list_targets and code_list_elements are redefined here in terms 
of indices into a vector of code_list_elements

***************************************************************************/
#include <util.h>
#include <input-control.h>
#include <quaternion-scalar.h>
#include <polynomial.h>
#include <braid-util.h>
#include <Gauss-code.h>
#include <reidemeister.h>

class code_list_element;

class list_target
{
public:

	/* location is the index in code_element of the code_list_element to which the list_target points */
	int location;  // location = -1 means that the code_list target takes us to the unlink
	
	vector<int> R1_move;
	matrix<int> R2_move; // two columns
	matrix<int> R3_move;  // 6-tuple <ingress over-arc-edge> <ingress under-arc-edge-1> <ingress over-arc-edge-2> <over-crossing-1> <over-crossing-2> <dominated-crossing>

	list_target(): location(-1) {}
	list_target(int l): location(l) {}
    friend ostream& operator << (ostream& outstream, list_target& t);    
    friend void write_list_target (ostream& outstream, list_target& t);    
    friend void print_list_target (ostream& outstream, list_target& t, string prefix);    
};

class code_list_element
{
public:
	int parent;  // used for building equivalence classes
	int num_crossings;
	int index; // the index of the code_list_element in code_element
	int class_identifier;  // true if this code_list_element has been included in an equivalence class
	int num_targets;
	string code;
	vector<list_target> target; 
	
	code_list_element(): parent(0), num_crossings(-1), index(-1), class_identifier(-1), num_targets(0) {}
    friend void print_code_list_element (ostream& outstream, code_list_element& e, string prefix);    
};

class equivalence_class
{
public:
	
	enum symmetry
	{
		none,
		line_reflect,
		plane_reflect,
		double_reflect,
	};
	
	string min_representation;
	int min_num_crossings;  // the number of crossings in the min_representation
	int root_index; // the index in code_element
	int class_identifier;
	
	/* symmetry_class stores the class_identifier of those symmetrical equivalence classes this class has replaced
	   
	   symmetry_type  indicates whether this equivalence class has been omitted due to finding a symmetrical class in the list.  
	   The value of none indicates that this class has been retained in the list of non-symmetrical classes 
	   
	   symmetry_representative stores the class identifier of the replacement class for this equivalence class if the symmetry_type
	   is not none
	*/
	list<int> symmetry_class; 
	int symmetry_type;  // an equivalence_class::symmetry value
	int symmetry_representative;  
	
	equivalence_class(): min_num_crossings(0),root_index(-1),class_identifier(-1),symmetry_type(equivalence_class::symmetry::none),symmetry_representative(-1){}
};

class Reidemeister_tree_node
{
public:
	
	int index; // the index of the code_list_element in code_element, a value of -1 indicates the unlink;
	int parent;  // the index of the code_list_element that is the parent of code_element[index] in the maximal tree
	int parent_tree_node_offset; // the index in tree_node of the Reidemeister_tree_node corresponding to the parent
	int root_hop_count;  // the number of list target hops to the root (i.e. the distance from the root of the tree)

	/* if parent_list_target is false, then root_list_target is the index of the list_target that points to the parent 
	   code_list_element in the maximal tree.
	   
	   if the Reidemeister_tree_node corresponds to the unlink, or contains no moves, parent_list_target is set true
	   and root_list_target corresponds to the index of the list target at the parent that points to this node
	*/
	bool parent_list_target;
	int root_list_target;  
	
	int num_R1_moves; // the cumulative number of Reidemeister 1 moves in the path to the root
	int num_R2_moves; // the cumulative number of Reidemeister 2 moves in the path to the root
	int num_R3_moves; // the cumulative number of Reidemeister 3 moves in the path to the root
	
	Reidemeister_tree_node(): index(-1),parent(-1),parent_tree_node_offset(-1),root_hop_count(0),parent_list_target(false),root_list_target(-1),num_R1_moves(0),num_R2_moves(0),num_R3_moves(0){}
    friend void print_Reidemeister_tree_node (ostream& outstream, Reidemeister_tree_node& n, string prefix);    
};


void print_code_list_element (ostream& outstream, code_list_element& e, string prefix="");
