/*************************************************************************
                      glist.h

***************************************************************************/
#include <util.h>
#include <input-control.h>
#include <Gauss-code.h>
#include <reidemeister.h>

class code_list_element;

class list_target
{
public:

	/* location == 0 means that the destination of the list_target is not a code_list_element on of the lists.  That is, the destination
	   is either the unlink or is a pair of disjoint codes each having a smaller number of components that currently being processed.
	   In the first case, the destination string is empty, indicating the unlink, in the latter case, the destination string holds the
	   two codes, concatenated with " && "
	   
	   In the event that location is non-zer0, the destination string is not used.
	*/
	code_list_element* location;  
	string destination;           
	
	list<int> R1_move;
	list<pair<int,int> > R2_move;
	list<vector<int> > R3_move;  // 6-tuple <ingress over-arc-edge> <ingress under-arc-edge-1> <ingress over-arc-edge-2> <over-crossing-1> <over-crossing-2> <dominated-crossing>

	list_target(): location(0) {}
	list_target(code_list_element* l): location(l) {}
	unsigned int num_moves() {return R1_move.size()+R2_move.size()+R3_move.size();}
    friend ostream& operator << (ostream& outstream, list_target& t);    
    friend void write_list_target (ostream& outstream, list_target& t);    
};

class code_list_element
{
public:
	code_list_element* parent;  // used for building equivalence classes root nodes have parent zero
	int num_crossings;
	int hash_index;  // the index of a code_list_element in the vector<list<code_list_element> > to which code_list_pointer[.] points
	int offset;
	int index; // the index of the preferred code within the set of <code_list index><list_offset> codes
	int member_equivalence_class;  // true if this code_list_element has been included in an equivalence class
	int num_targets;
	string code;
	vector<list_target> target; 
	
	code_list_element(): parent(0), num_crossings(-1), hash_index(-1), offset(-1), index(-1), member_equivalence_class(0), num_targets(0) {}
	code_list_element(int n, int i, int o): parent(0),  num_crossings(n), hash_index(i), offset(o), index(-1), member_equivalence_class(0), num_targets(0) {}
    friend void print_code_list_element (ostream& outstream, code_list_element& e, string prefix);    
    friend void write_code_list_element (ostream& outstream, code_list_element& e,int class_index);    
};

class graft_data
{
public:
	code_list_element* from_node;
	code_list_element* to_node;
	
	graft_data(): from_node(0),to_node(0){}
	graft_data(code_list_element* f, code_list_element* t): from_node(f),to_node(t){}
};

class equivalence_class
{
public:
	enum symmetry
	{
		none,
		line_reflect,
		plane_reflect,
		double_reflect,
	};

	string min_representation;
	int min_num_crossings;  // the number of crossings in the min_representation, zero if the class is degenerate
	code_list_element* root;
	list<int> class_identifiers;
	
	/* symmetry_class stores the class_identifier of those symmetrical equivalence classes this class has replaced
	   
	   symmetry_type  indicates whether this equivalence class has been omitted due to finding a symmetrical class in the list.  
	   The value of none indicates that this class has been retained in the list of non-symmetrical classes 
	   
	   symmetry_representative stores the class identifier of the replacement class for this equivalence class if the symmetry_type
	   is not none
	*/	
	list<int> symmetry_class; 
	int symmetry_type;  // an equivalence_class::symmetry value
	int symmetry_representative;  
	
	equivalence_class(): min_representation(""),min_num_crossings(0),root(0),symmetry_type(equivalence_class::symmetry::none),symmetry_representative(-1){}
	
	bool change_to_representation(string& s);
};

/* function object for comparing code strings with 'find_if' */
struct code_equals
{
	string code;
	code_equals(string s): code(s) {}
	bool operator () (code_list_element& e) {return e.code == code;}
};

/* function object for comparing root location with 'find_if' */
struct class_identifier_includes
{
	int class_identifier;
	class_identifier_includes(int c): class_identifier(c) {}
	bool operator () (const equivalence_class& ec) {return (find(ec.class_identifiers.begin(),ec.class_identifiers.end(),class_identifier) != ec.class_identifiers.end());}
};

/* function object for comparing index values with 'find_if' */
struct index_equals
{
	int index;
	index_equals(int i): index(i) {}
	bool operator () (const code_list_element* e) {return e->index == index;}
};

void print_code_list_element (ostream& outstream, code_list_element& e, string prefix="");    

