/************************************************************************

**************************************************************************/
using namespace std;
#include <fstream>
#include <iostream>
#include <sstream>
#include <cstdlib>
#include <cstring>
#include <vector>
#include <list>
#include <algorithm>
#include <iomanip>

/********************* External variables ***********************/
extern bool			UNORIENTED_SEARCH;
extern bool			WRITE_ELEMENT_POINTERS;
extern ofstream     debug;
extern ofstream     output;
extern int			max_num_crossings;
extern int			num_components;
extern int 			hash_threshold;
extern int 			shift_power;
extern string		filepath;
extern bool     	DEVELOPMENT_MODE;



#include <Gauss-code.h>

#define ONE_ROW_AND_COLUMN		1

/********************* Function prototypes ************************/

Gauss_code::Gauss_code(string s, int num_terms, int num_components)
{
	istringstream iss(s);
	code = vector<code_term_type>(num_terms);
	num_terms_in_component = vector<int>(num_components);
	
	char c;
	int crossing;
	int component = 0;
	int term_count = 0;
	int index=-1;
	
	while(iss >> c)
	{
		if (c == 'O')
		{
			index++;
			iss >> crossing;			
			code[index] = 6*crossing-5; // 6*(crossing-1)+1
			term_count++;	
		}
		else if (c == 'U')
		{
			index++;
			iss >> crossing;
			code[index] = 6*crossing-2; // 6*(crossing-1)+4
			term_count++;
		}
		else if (c == ',')
		{
			num_terms_in_component[component] = term_count;

			component++;
			term_count = 0;
			
		}
		else if (c == '+')
		{
			code[index]++;
		}
		else if (c == '-')
		{
			code[index]--;
		}
		else
		{
			cout << "Gauss_code constructor read spurious character " << c << " from " << s << endl;
			exit(0);
		}
	}	

	num_terms_in_component[component] = term_count;	
}


Gauss_code::Gauss_code(generic_code_data code_data)
{
	gauss_orientation_data gdata(code_data);
	num_terms_in_component = gdata.num_terms_in_component;
	code = vector<int>(gdata.num_terms);
	
	for (int i=0; i< gdata.num_terms; i++)
	{
		int crossing = abs(gdata.classical_gauss_data[i])-1;
		code[i] = 6*(crossing) + (gdata.classical_gauss_data[i]>0?1:4) + gdata.classical_crossing_sign[crossing];
	}
	
}	


void write_gauss_code(ostream& os, Gauss_code& g_code, int n, bool raw_format)
{
	vector<code_term_type>& c = g_code.code;
	
	int limit = (n? n: c.size());
	
	int index = 0;
	for (size_t i=0; i< g_code.num_terms_in_component.size() && index < limit; i++)
	{
		for (int j = 0; j < g_code.num_terms_in_component[i] && index < limit; j++) 
		{
			int crossing_detail = c[index]%6;
			os << (crossing_detail/3 == 1 ? "U" : "O");
			os << c[index]/6 + 1;
			if (crossing_detail%3 == 0)
				os << '-';
			else if (crossing_detail%3 == 2)
				os << '+';
			index++;
		}	
		
		if (i< g_code.num_terms_in_component.size()-1 && index < limit)
			os << ',';
	}

	if (raw_format)
	{
		for (int i = 0; i < limit; i++) 
		{
			os << ' ' << c[i];
		}
	}
}	  


void print_gauss_code(ostream& os, Gauss_code& g_code)
{	
	if (g_code.num_terms_in_component.size() > 1)
	{
		for (size_t i = 0; i < g_code.num_terms_in_component.size(); i++) 
			os << g_code.num_terms_in_component[i] << ' ';
	}
		
	for (size_t i=0; i< g_code.code.size(); i++)
	{
		os << g_code.code[i];
		if (i < g_code.code.size()-1)
			os << ' ';
	}
}	  
  

void write_signed_code(ostream& os, Gauss_code& g_code)
{
	int num_code_terms = g_code.code.size();
	int num_crossings = num_code_terms/2;

	int num_signs = 1 << num_crossings;
	for (int i=0; i< num_signs; i++)
	{
		vector<int> sign(num_crossings);
		for (int j=0; j<num_crossings; j++)
		{
			if (i & (1 << j) ) // if the jth bit of i is set it's a positive crossing otherwise negative
				sign[num_crossings-j-1] = 1;
			else
				sign[num_crossings-j-1] = -1;
			
		}

		Gauss_code signed_code=g_code;
		for (int j=0; j< num_code_terms; j++)
			signed_code.code[j] += sign[signed_code.code[j]/6];

		write_gauss_code(os, signed_code);
	}
}	  

/* minimal_cyclic_code is first used by jglist to enumerate unsigned Gauss codes; that is sequences of <O|U><crossing-number> with a fixed number of terms in each component.  It
   returns a minimal equivalent code to g_code based on a vector comparison of the encoding 6*<crossing_number>+<crossing_detail>, where <crossing_detail>=enum{O-, O, O+, U-, U, U+}.   
   To determine the minimal code, the function cycles each component in all possible ways and, if UNORIENTED_SEARCH is true, reverses the component orientation in all possible ways but 
   does *not* consider re-ordering of components in order to be consistent with the enumeration in jglist.  For each starting point and cycle, the function re-numbers the crossings in 
   the order in which they are encountered and makes a vector comparison that is independent of component boundaries, since the component order is fixed.
  
   Every Gauss code for a given diagram has a corresponding unsigned code obtained by ignoring the sign of each crossing and therefore every diagram has a minimal unsigned code.  However, 
   there may be multiple component starting points and orientations that give rise to the same minimal code.  Thus jglist determines all minimal unsigned codes and then assigns crossing 
   signs in all possible ways.
   
   mimimal_cyclic_code is also used by jgprocess to determine which Gauss code to look for when computing list targets corresponding to Reidemeister II or Reidemeister III moves.  
   For this use case, the Gauss codes are signed.  Since minimal_cyclic_code isolates the <crossing_detail> before a crossing is renumbered and then re-applies the crossing detail, there
   are no differences in the funtion when it is evauating minimal signed or unsigned codes.
   
   Note, however, that the signed codes produced by the jglist algorithm by allocating crossing signs to minimal unsigned Gauss codes does not produce minimal signed Gauss codes.  An example is
   the minimal unsigned code O1O2U1U2: if crossing 1 is positive and crossing 2 negative, the jglist algorithm gives us O1+O2-U1+U2- which has minimal signed code 01-U2+U1-O2+.  This is because 
   the encoding for O1+ is 2 whereas that for O1- is 1.  For this reason, minimal versions of the signed codes produced by jglist are calculated in jgprocess.
   
   Note also that if two distinct unsigned codes U_1 and U_2 are allocated a set of signs S_1 and S_2, then the minimal signed code for {U_1,S_1} and {U_2,S_2} are distinct.  This is because, if they were 
   the same, there would be a diagram admitting both {U_1,S_1} and {U_2,S_2} as Gauss codes.  That means that U_1 and U_2 must be the same.
   
   Note also that the approach of numbering crossings in the order in which they are encountered produces more "minimal" codes than does the over-preffered approach.  The jglist method produces four 
   minimal signed codes with one component and two crossings: 
   
    O1-O2-U1-U2-
    O1-O2+U1-U2+
    O1-U2+U1-O2+
    O1+O2+U1+U2+

   whereas the over-preferred approach procudes three codes

	1 2 -1 -2/+ +
	1 2 -1 -2/+ -
	1 2 -1 -2/- -
   
   since the over-preferred code for O1-U2+U1-O2+ is 1 2 -1 -2/+ -
*/
vector<code_term_type> minimal_cyclic_code(Gauss_code& g_code)
{
	int num_code_terms = g_code.code.size();
	int num_crossings = num_code_terms/2;
	int num_components = g_code.num_terms_in_component.size();
	
if (debug_control::DEBUG >= debug_control::DETAIL)
{
	debug << "minimal_cyclic_code: presented with code: ";
	print_gauss_code(debug,g_code);	
	debug << ": ";
	write_gauss_code(debug,g_code);
	debug << endl;
}
	vector<code_term_type> minimal_code(num_code_terms, 6*num_crossings);	// bigger than any code we will be considering
	vector<int>& num_terms_in_component = g_code.num_terms_in_component;	
	vector<bool> r_flag(num_components); // initializes to false
	bool finished = false;
	
	do
	{
		/* consider the orientation determined by the r_flags */
		
if (debug_control::DEBUG >= debug_control::EXHAUSTIVE)
{
	debug << "minimal_cyclic_code: reverse flags: ";
	for (int i=0; i< num_components; i++)
		debug << r_flag[i] << " ";
	debug << endl;
}
				
		/* component_cycle will record the cyclic rotation of each component by recording at which term of the component 
		   we are going to start.
		*/
		vector<int> component_cycle(num_components); // initializes to zero

		bool complete = false;		
		do
		{
if (debug_control::DEBUG >= debug_control::EXHAUSTIVE)
{
	debug << "minimal_cyclic_code:     component_cycle = ";	
	for (int i=0; i< num_components; i++)
		debug << component_cycle[i] << ' ';
	debug << endl;
}	
			int start_of_component = 0;
			int next_new_crossing_number = 0;
			vector<code_term_type> rotation_code(num_code_terms);
			vector<int> crossing_perm(num_code_terms,-1);

			
			for (int component=0; component< num_components; component++)
			{
				int num_component_terms = num_terms_in_component[component];

if (debug_control::DEBUG >= debug_control::EXHAUSTIVE)
{
	debug << "minimal_cyclic_code:       component = " << component << ", num_component_terms = " << num_component_terms 
	      << ", start_of_component = " << start_of_component << endl;
}	
					
				for (int i=0; i < num_component_terms; i++)
				{
					int next_term;
					if (r_flag[component])
						next_term = start_of_component + (component_cycle[component]-i+num_component_terms)%num_component_terms;
					else
						next_term = start_of_component + (component_cycle[component]+i)%num_component_terms;
					
if (debug_control::DEBUG >= debug_control::EXHAUSTIVE)
	debug << "minimal_cyclic_code:         next_term " << next_term;
									
					int old_crossing = g_code.code[next_term]/6;
					int crossing_detail = g_code.code[next_term]%6;

if (debug_control::DEBUG >= debug_control::EXHAUSTIVE)
	debug << ", old_crossing = " << old_crossing << ", crossing_detail = " << crossing_detail << endl;
		
					if (crossing_perm[old_crossing] == -1)
					{
if (debug_control::DEBUG >= debug_control::EXHAUSTIVE)
	debug << "minimal_cyclic_code:         crossing " << old_crossing << " becomes new crossing " << next_new_crossing_number << endl;

						crossing_perm[old_crossing] = next_new_crossing_number;
						next_new_crossing_number++;				 
					}

					rotation_code[start_of_component+i] = 6*crossing_perm[old_crossing]+crossing_detail;
				}
				
				start_of_component += num_component_terms;
			}

if (debug_control::DEBUG >= debug_control::EXHAUSTIVE)
{
	debug << "minimal_cyclic_code:     rotation code";	
	int index=0;
	for (int i=0; i< num_components; i++)
	{
		for (int j=0; j < num_terms_in_component[i]; j++)
			debug << ' ' << rotation_code[index++];
			
		if (i< num_components-1)
			debug << ',';
	}
}

			if (rotation_code < minimal_code)
			{
				minimal_code = rotation_code;	
					
if (debug_control::DEBUG >= debug_control::EXHAUSTIVE)
	debug << " is preferred" << endl;
			}
			else
			{
if (debug_control::DEBUG >= debug_control::EXHAUSTIVE)
	debug << " not preferred" << endl;
			}

			/* increment component_cycle lexicographically up to the number of terms in each component.	*/
			for (int i= num_components-1; i >=0; i--)
			{
				component_cycle[i]++;
				
				if (component_cycle[i] >= num_terms_in_component[i])
				{
					component_cycle[i] = 0;
					if (i== 0)
					{
						complete = true;
						break;
					}
				}
				else
					break;
			}
		} while (!complete);
					
		/* if we're dealing with an unoriented Gauss code, consider other orientations of the components */
		if (UNORIENTED_SEARCH)
		{			
			finished = true;
			for (int i=num_components-1; i>=0; i--)
			{
				if (r_flag[i])
					r_flag[i] = false;
				else
				{
					r_flag[i] = true;
					finished = false;
					break;
				}
			}												
		}
		else
		{
			break;
		}		
		
	} while (!finished);
				
if (debug_control::DEBUG >= debug_control::DETAIL)
{
	debug << "minimal_cyclic_code: returning minimal code: ";
	for (int i=0; i< num_code_terms; i++)
		debug << minimal_code[i] << ' ';
	debug << endl;
	
}

	return minimal_code;	
}
