/**************************************************************************************************************************
ganalyse provides analytical tools for exploring the output of gprocess

1. Remove all line plane and double reflections together with their inverses (in the unoriented case) from the equivalence 
   classes and write the minimum representative of each class to output.

April 2022

***************************************************************************************************************************/
using namespace std;
#include <fstream>
#include <iostream>
#include <sstream>
#include <cstdlib>
#include <cstring>
#include <vector>
#include <list>
#include <algorithm>
#include <csignal>
#include <iomanip>
#include <ctype.h>
#include <functional>

/******************** Global Variables **********************/
bool 			UNORIENTED_SEARCH = false;
bool			EQUIVALENCE_CLASS_DETAIL = false;
bool			DISTINCT_NON_TRIVIAL_CLASSES = false;
bool			COMPARE_CODES = false;
bool			REIDEMEISTER_MOVES = false;
bool			sigusr1_request = false;
bool			termination_request = false;
int				max_num_crossings=0;
int 			hash_threshold=100;
int 			shift_power = 23;
string			filepath;

bool     		DEVELOPMENT_MODE = false; 

ofstream 	output;
extern ofstream    debug;

#include <ganalyse.h>

#define UNLINK -1

/********************* Function prototypes ***********************/
void set_programme_long_option(char* cptr, string source);
void set_programme_short_option(char* cptr);
void print_prog_params (ostream& os, string level, string prefix);
void check_debug_option_parameters(char* pptr, string option);
bool debug_setup(char* argv_ptr);
void debug_help();
void helpinfo();
void parse_equivalence_class(vector<equivalence_class>& code_class, string next_line);
void parse_code_list_element(vector<equivalence_class>& code_class, vector<code_list_element>& code_element, string next_line);
void parse_symmetry_data(vector<equivalence_class>& code_class, string next_line);
void write_equivalence_classes(vector<equivalence_class>& code_class, vector<code_list_element>& code_element, ofstream& os);
void write_code_list_element (ostream& outstream, code_list_element& e);
void write_list_target(ostream& outstream, list_target& t);
string gauss_line_reflect (string c, bool unoriented);
string gauss_plane_reflect (string c, bool unoriented);
string minimum_class_representative(vector<list<int> >& code_hash_list, vector<equivalence_class>& code_class, vector<code_list_element>& code_element, string s);
int code_element_index(vector<list<int> >& code_hash_list, vector<code_list_element>& code_element, string s);
vector<int> count_equivalence_class_members(vector<code_list_element>& code_element, int num_code_list_elements, int num_equivalence_classes);
int tree_node_index(vector<Reidemeister_tree_node>& tree_node,int node_count, int index);
bool preferred_parent(Reidemeister_tree_node& next_tree_node, list_target& neighbour_target, Reidemeister_tree_node& neighbour_tree_node);
void print_Reidemeister_moves(ostream& outstream, list_target& target);
void print_equivalence_class(equivalence_class& e, ofstream& os, string prefix);
void describe_reidemeister_moves(vector<code_list_element>& code_element,vector<int>& class_element_count, int index_1, int index_2, string special_case_description="unknot");
namespace util
{
	string itos(long n);
}

void sigterm_handler (int sig) // kill -15, or just kill
{
	cout << "\ngprocess programme received SIGTERM, setting termination_request=true" << endl;
	termination_request = true;
}

/* sigusr1_handler causes gprocess to report progress on the current task and continue. */
void sigusr1_handler (int sig) // kill -10
{
	cout << "\ngprocess programme received SIGUSR1, setting sigusr1_request=true" << endl;
	sigusr1_request = true;
}



/******************* Main Function ************************/

int main (int argc, char* argv[])
{	
	signal(SIGTERM, sigterm_handler); // register the SIGTERM handler
	signal(SIGUSR1, sigusr1_handler); // register the SIGUSR1 handler
	
	/* take a copy of the original command line in case we're debugging and remove
	   the debug details.  We will want to record the original command line for the debug file.
	*/
	
	vector<string> original_programme_arguments(argc); 
	for (int i = 0; i< argc; i++)
		original_programme_arguments[i] = string(argv[i]);
	
	string input_file;
	ofstream tempfile;

	time_t start_time = time(0);
	std::tm* start_timeinfo = localtime (&start_time);
	
	if (argc == 1)
	{
		helpinfo();
		exit(0);
	}
	
    /* Read the programme parameters and input files */
	for (int i=1; i < argc; i++)
	{
		if (*(argv[i]) == '-')
		{
			if (*(argv[i]+1) == '-')
				set_programme_long_option(argv[i], "command line");
			else 
				set_programme_short_option(argv[i]);
		}
		else if (!input_file.length()) 
		{
			input_file = argv[i];
		}
	}	
	
if (debug_control::DEBUG >= debug_control::SUMMARY)
	print_prog_params (debug, "basic", "");

if (debug_control::DEBUG >= debug_control::SUMMARY)
{
	debug << "full command line: ";
	for (int i = 0; i< argc; i++)
		debug << original_programme_arguments[i] << ' ';
	debug << endl;
}


/*DEVELOPMENT_MODE is used for testing individual functions during development */
if (DEVELOPMENT_MODE)
{	
	exit(0);
}
/* End of DEVELOPMENT_MODE code */
		
    output.open ("ganalyse.out");

    if (!output)
    {
		cout << "\nError opening output file\n";
		exit(0);
	}
	else
	{
		output << "Output from ganalyse ";
		for (int i=1; i< argc; i++	)
			output << argv[i] << " ";
		output << "\n";
		
		output << "Start time: " << asctime(start_timeinfo);
	}
	
	ifstream input(input_file.c_str());
	
	if(!input)
	{
		cout << "\nError opening input file " << input_file << endl;
		exit(0);
	}	
	else
	{
if (debug_control::DEBUG >= debug_control::DETAIL)
	debug << "ganalyse: reading equivalence class data from " << input_file << " file" << endl;
	}

	int num_code_list_elements=0;
	int num_equivalence_classes = 0;
	string next_line;
	while(getline(input,next_line))
	{
		
		if (next_line.find("CLASSCOUNT") != string::npos)
		{
			char c;
			istringstream iss(next_line);
			do {iss >> c;} while (c != ':');
			iss  >> num_equivalence_classes;

	cout << "\n" << input_file << " contains " << num_equivalence_classes << " equivalence classes" << endl;

if (debug_control::DEBUG >= debug_control::SUMMARY)
	debug << "ganalyse: " << input_file << " contains " << num_equivalence_classes << " equivalence classes" << endl;
		
		}
		else if (next_line.find("NODECOUNT") != string::npos)
		{
			char c;
			istringstream iss(next_line);
			do {iss >> c;} while (c != ':');
			iss  >> num_code_list_elements;

if (debug_control::DEBUG >= debug_control::SUMMARY)
	debug << "ganalyse: " << input_file << " contains " << num_code_list_elements << " code list elements" << endl;
		
			break;
		}
	}

	vector<code_list_element> code_element(num_code_list_elements);
	
	vector<equivalence_class> code_class(num_equivalence_classes);
	
	/* read the equivalence classes from the input file, which has a comma delimited format.  
	
		class:			
		if symmetry-type == none	   
		   <-1>,<min-gauss-code-representation>,<class-identifier>:<symmetry-type>:<symmetry-class-length>:<symmetry-class>
		if symmetry-type != none	   
		   <-1>,<min-gauss-code-representation>,<class-identifier>:<symmetry-type>:<symmetry_representative>:<symmetry-class-length>:<symmetry-class>
	   
		where <symmetry-class> = <element_first>:...:<element_last>
		 
		node:  <index>,<parent-index>,<gauss-code>,<class-identifier>,<num-list-targets>,{<list-target>}

		where: <parent-index> is set to -1 if the node is the root of a class-tree
		       {<class-identifiers>} is a comma delimited set of integers
		       {<list-target>} is a comma delimted set of list-targets.
		       
		Each <list-target> has the form:
		
		<index>,<num-R1-moves>,<num-R2-moves>,<num-R3-moves>,{<R1-move>},{<R2-move>},{<R3-move>}
		
		where: <index> is set to -1 if the list target location is zero
		       <R1-move> = list of integers
		       <R2-move> = list of integer pairs
		       <R3-move> = list of integer triples (<over-crossing-1> <over-crossing-2> <dominated-crossing>)			
	*/	

	while(getline(input,next_line))
	{		
if (debug_control::DEBUG >= debug_control::DETAIL)
	debug << "ganalyse: next_line = " << next_line << endl;
	
		if (next_line[0] == '-')
			parse_equivalence_class(code_class, next_line);
		else if (isdigit(next_line[0]))
			parse_code_list_element(code_class, code_element, next_line);
	}

	vector<int> class_element_count = count_equivalence_class_members(code_element, num_code_list_elements, num_equivalence_classes); 

if (debug_control::DEBUG >= debug_control::DETAIL)
{
	debug << "ganalyse: class_element_count: ";
	int check=0;
	for (int i=0; i< num_equivalence_classes; i++)
	{
		check+= class_element_count[i];
		debug << class_element_count[i] << ' ';
	}
	debug << "\nganalyse: check total = "  << check << endl;
}

	/* in order to speed up searching for codes in code_element, we create a vector of lists of indices into code_element, 
	   indexed by the code hash, so we need only check a few elements of code_element when looking for a code
	*/
	int code_list_length = 1 << shift_power;
	vector<list<int> > code_hash_list(code_list_length);  
	for (int i=0; i< num_code_list_elements; i++)
	{
		size_t hash_val = hash<string>()(code_element[i].code);
		hash_val >>= 64-shift_power; // the hash is 64 bits long
		code_hash_list[hash_val].push_back(code_element[i].index);

if (debug_control::DEBUG >= debug_control::DETAIL)
	debug << "code string " << code_element[i].code << ", hash_val = " << hash_val << ", index = " << code_element[i].index << endl;
	}
	
	cout << "created code hash list" << endl;

if (debug_control::DEBUG >= debug_control::DETAIL)
{
	debug << "ganalyse: code_hash_list: " << endl;
	for (int i=0; i< code_list_length; i++)
	{
		if (code_hash_list[i].size())
		{
			debug << "  index " << i << ": ";
			list<int>::iterator lptr = code_hash_list[i].begin();
			while (lptr != code_hash_list[i].end())
			{
				debug << *lptr << ' ';
				lptr++;
			}
			debug << endl;
		}
	}
}
	
	if (EQUIVALENCE_CLASS_DETAIL)
		write_equivalence_classes(code_class,code_element,output);

	if (DISTINCT_NON_TRIVIAL_CLASSES)
	{
		for (int i=0; i < num_equivalence_classes; i++)	
		{
			if (code_class[i].min_representation == "E" || code_class[i].min_representation.find(" && ") != string::npos)
				continue;

			if (code_class[i].symmetry_type == equivalence_class::symmetry::none)
				output << code_class[i].min_representation << endl;
		}

	}	
	else if (COMPARE_CODES || REIDEMEISTER_MOVES)
	{
		bool finished = false;
		do
		{
			
			/* Read a pair of codes from cin and identify their corresponding code_list_elements */
			cout << "\nEnter first Gauss code: ";
			string input_1;
			getline(cin,input_1);
			
			if (input_1 == "" || input_1 == "q" || input_1 == "exit")
			{
				finished = true;
				continue;
			}
			
			cout << "Enter second Gauss code: ";
			string input_2;
			getline(cin,input_2);
		
if (debug_control::DEBUG >= debug_control::SUMMARY)
	debug << "ganalyse: read codes input_1 = " << input_1 << ", input_2 = " << input_2 << endl;
		
			generic_code_data gdata_1;
			read_gauss_code(gdata_1,input_1);
			generic_code_data gdata_2;
			read_gauss_code(gdata_2,input_2);

			string preferred_1 = over_preferred_gauss_code(gdata_1,UNORIENTED_SEARCH);
			string preferred_2 = over_preferred_gauss_code(gdata_2,UNORIENTED_SEARCH);
			
			output << "\nFirst input Gauss code: " << input_1 << ", over-preferred form " << preferred_1 << endl;
			output << "\nSecond input Gauss code: " << input_2 << ", over-preferred form " << preferred_2 << endl;
	
			int index_1 = code_element_index(code_hash_list, code_element, preferred_1);
			int index_2 = code_element_index(code_hash_list, code_element, preferred_2);

			if (index_1 == index_2)
			{
				cout << "Codes are identical" << endl;
				output << "  Codes are identical" << endl;
			}
			else if (code_element[index_1].class_identifier == code_element[index_2].class_identifier)
			{
				/* includes the case where the two codes are part of the same trivial equivalence class */
				if (COMPARE_CODES)
				{
					cout << "Codes are Reidemeister equivalent" << endl;
				}
				else if (REIDEMEISTER_MOVES)
					describe_reidemeister_moves(code_element,class_element_count,index_1,index_2);		
			}
			else 
			{
				int start_class_id;
				int end_class_id;
				bool trace_symmetry = false;
				int class_1 = code_element[index_1].class_identifier;
				int class_2 = code_element[index_2].class_identifier;

if (debug_control::DEBUG >= debug_control::SUMMARY)
{
	debug << "ganalyse: class " << class_1 << endl;
	print_equivalence_class(code_class[class_1],debug,"ganalyse:  ");
	debug << "ganalyse: class " << class_2 << endl;
	print_equivalence_class(code_class[class_2],debug,"ganalyse:  ");
}

				if (code_class[class_1].min_representation == "E" && code_class[class_2].min_representation == "E")
				{
					// both codes representations of the unknot, find moves from each code to the unknot
					cout << "Both codes are Reidemeister equivalent to the unlink.\nReidemeister moves for first code:" << endl;
					output << "  Both codes are Reidemeister equivalent to the unlink.\nReidemeister moves for first code:" << endl;
					describe_reidemeister_moves(code_element,class_element_count,index_1,-1);
					cout << "Reidemeister moves for second code:" << endl;
					output << "  Reidemeister moves for second code:" << endl;
					describe_reidemeister_moves(code_element,class_element_count,index_2,-1);
				}
				else if (code_class[class_1].min_representation.find(" && ") != string::npos || code_class[class_2].min_representation.find(" && ") != string::npos)
				{
					/* if one or both codes are equivalent to a disjoint concatenation of codes, leave how to proceed to human interpretation, 
					   just provide moves from each such code to the disconnected state
					*/
					if(code_class[class_1].min_representation.find(" && ") != string::npos)
					{
						cout << "First code is Reidemeister equivalent to the disjoint sum of codes " << code_class[class_1].min_representation << endl;
						output << "  First code is Reidemeister equivalent to the disjoint sum of codes " << code_class[class_1].min_representation << endl;
						describe_reidemeister_moves(code_element,class_element_count,index_1,-1,"disconnected diagram");
					}
					
					if(code_class[class_2].min_representation.find(" && ") != string::npos)
					{
						cout << "Second code is Reidemeister equivalent to the disjoint sum of codes " << code_class[class_2].min_representation << endl;
						output << "  Second code is Reidemeister equivalent to the disjoint sum of codes " << code_class[class_2].min_representation << endl;
						describe_reidemeister_moves(code_element,class_element_count,index_2,-1, "disconnected diagram");
					}
				}
				else
				{
					cout << "Codes are not Reidemeister equivalent" << endl;
					output << "  Codes are not Reidemeister equivalent" << endl;

					if (find(code_class[class_1].symmetry_class.begin(),code_class[class_1].symmetry_class.end(), class_2) != code_class[class_1].symmetry_class.end()) 
					{
						start_class_id = class_2;
						end_class_id = class_1;
						trace_symmetry = true;
					}
					else if (find(code_class[class_2].symmetry_class.begin(),code_class[class_2].symmetry_class.end(), class_1) != code_class[class_2].symmetry_class.end())
					{
						start_class_id = class_1;
						end_class_id = class_2;
						trace_symmetry = true;
					}
	
					int class_id = start_class_id;
					
					if (trace_symmetry)
					{
						bool reached_other_code = false;
						int cumulative_line_symmetries = 0;
						int cumulative_plane_symmetries = 0;
						
						cout << code_class[class_id].min_representation << " is the ";
						output << "  " << code_class[class_id].min_representation << " is the ";
						
						do
						{
						
if (debug_control::DEBUG >= debug_control::SUMMARY)
	debug << "ganalyse: " << code_class[class_id].min_representation << " is the ";

							switch(code_class[class_id].symmetry_type)
							{
								case equivalence_class::symmetry::line_reflect: 
									cumulative_line_symmetries++; 
if (debug_control::DEBUG >= debug_control::SUMMARY)
	debug << "line"; 
									break; 
								case equivalence_class::symmetry::plane_reflect: 
									cumulative_plane_symmetries++; 
if (debug_control::DEBUG >= debug_control::SUMMARY)
	debug << "plane"; 							
									break; 
								case equivalence_class::symmetry::double_reflect: 
									cumulative_line_symmetries++; 
									cumulative_plane_symmetries++; 
if (debug_control::DEBUG >= debug_control::SUMMARY)
	debug << "double"; 							
									break; 
								default: 
if (debug_control::DEBUG >= debug_control::SUMMARY)
	debug << "UNKNOWN"; 							
							
							}						

//					cout  << " reflection of " << code_class[code_class[class_id].symmetry_representative].min_representation << endl;
if (debug_control::DEBUG >= debug_control::SUMMARY)
	debug  << " reflection of " << code_class[code_class[class_id].symmetry_representative].min_representation << endl;

							if (code_class[class_id].symmetry_representative == end_class_id)
								reached_other_code = true;
							else
								class_id = code_class[class_id].symmetry_representative;
		
						} while (!reached_other_code);
		
						cumulative_line_symmetries %= 2; 
						cumulative_plane_symmetries %= 2;

if (debug_control::DEBUG >= debug_control::SUMMARY)
	debug  << " cumulative reflections results in " << cumulative_line_symmetries << " line and " << cumulative_plane_symmetries << " plane symmetries" << endl;
				
						if (cumulative_line_symmetries & cumulative_plane_symmetries)
						{
							cout << "line and plane reflection of ";
							output << "line and plane reflection of ";
						}
						else if (cumulative_line_symmetries)
						{
							cout << "line reflection of ";
							output << "line reflection of ";
						}
						else if (cumulative_plane_symmetries)
						{
							cout << "plane reflection of ";
							output << "plane reflection of ";
						}
						
						cout << code_class[end_class_id].min_representation << endl;
						output << code_class[end_class_id].min_representation << endl;
										
						if (cumulative_line_symmetries + cumulative_plane_symmetries != 0)
						{
							string reflected_code = code_element[(start_class_id == class_1?index_1:index_2)].code;
	
if (debug_control::DEBUG >= debug_control::SUMMARY)
		debug << "ganalyse: reflecting code " << reflected_code << endl;
		
							if (cumulative_line_symmetries != 0)
							{
								reflected_code = gauss_line_reflect(reflected_code,UNORIENTED_SEARCH);
if (debug_control::DEBUG >= debug_control::SUMMARY)
		debug << "ganalyse: line reflection gives code " << reflected_code << endl;
							}
								
							if (cumulative_plane_symmetries != 0)
							{
								reflected_code = gauss_plane_reflect(reflected_code,UNORIENTED_SEARCH);
if (debug_control::DEBUG >= debug_control::SUMMARY)
		debug << "ganalyse: plane reflection gives code " << reflected_code << endl;
							}
		
							if (start_class_id == class_1)
								index_1 = code_element_index(code_hash_list, code_element, reflected_code);
							else
								index_2 = code_element_index(code_hash_list, code_element, reflected_code);
if (debug_control::DEBUG >= debug_control::SUMMARY)
{
		debug << "ganalyse: new code_element indices: index_1 = " << index_1 << ", class_identifier " << code_element[index_1].class_identifier 
		                                        << ", index_2 = " << index_2 << ", class_identifier " << code_element[index_2].class_identifier << endl;
}		
						
						}
		
						if (code_element[index_1].class_identifier != code_element[index_2].class_identifier)
						{
if (debug_control::DEBUG >= debug_control::SUMMARY)
	debug << "ganalyse: Error! Symmetry check does not produce codes in the smae equivalence class" << endl;
						cout << "Error! Symmetry check does not produce codes in the smae equivalence class";
						exit(0);
						}
	
						if (index_1 == index_2)
						{
							cout << "No further Reidemeister moves required" << endl;
if (debug_control::DEBUG >= debug_control::SUMMARY)
	debug << "ganalyse: no reflection symmetries found" << endl;
						}
						else
						{
							describe_reidemeister_moves(code_element,class_element_count,index_1,index_2);
						}				
					}
					else
					{
						cout << "Codes are not related by reflections." << endl;
						output << "  Codes are not related by reflections." << endl;
					}
				}
			}
		} while(!finished);	
	}
	
	time_t end_time = time(0);
	double time_exp = difftime(end_time, start_time);
	int dd = static_cast<int>(time_exp/3600/24);
	time_exp -= dd*3600*24;
	int hh = static_cast<int>(time_exp/3600);
	time_exp -= hh*3600;
	int mm = static_cast<int>(time_exp/60);
	time_exp -= mm*60;
	int ss = static_cast<int>(time_exp);
	output << "\nRunning time: " << dd << " days " << hh << " hours " << mm	<< " minutes " << ss << " seconds" << endl;

	output.close();
	cout << endl;

	return 0;
} /* End of main */

/***************  Functions required for setting up gprocess specific debug **************/

void set_main_debug_option_parameter(char* pptr, string option);
void set_main_debug_option(char* start, char* end)
{
	char  loc_buf[end-start+2];
	char* c1 = start;
	char* c2 = loc_buf;

	/* if both start and end are zero, display debug help information.
	   this has been included here in this manner so that each time a debug option
	   is added, the help will be updated (hopefully!)
	*/
	if (start == 0 && end == 0)
	{
		set_main_debug_option_parameter(0,"ganalyse");
		return;
	}

	do
	{
		*c2++ = *c1++;
	} while (c1 <= end);
	
	*c2 = '\0';

	char* pptr = strchr(loc_buf,'{');
	if (pptr)
	{
		*pptr++ = '\0';
	}

	/* now, even if there are parameters, the first part of loc_buf 
	   is a C-string that identifies the option */
	
	if (!strcmp(loc_buf,"ganalyse"))
	{
		debug_control::DEBUG = debug_control::SUMMARY;
		debug << "set_main_debug_option: setting debug option debug_control::DEBUG = debug_control::SUMMARY" << endl;		
		
		if (pptr)
		{
			/* check for any parameters */
			check_debug_option_parameters(pptr, "ganalyse");
		}
	}

	debug.flush();

}

void helpinfo()
{
	cout << "\nUsage: ganalyse [-<short-option>] [--<long-option>] <input-file>";
	cout << "\n<short-option>:";
	cout << "\n       #: debug, see #H! for details";
	cout << "\n       c: write non-trivial distinct equivalence classes";
	cout << "\n       e: write full equivalence class detail";
	cout << "\n       h: show normal help";
	cout << "\n       u: unoriented code lists";
	cout << "\n       H!: show this help";
	cout << "\n       #H!: display debug help screen";
	cout << "\n<long-option>:";
	cout << "\n       compare: compare Gauss codes";
	cout << "\n       moves: evaluate a sequence of Reidemeister moves between two Gauss codes";
	cout << "\n<input-file>: relative path to a file containing the output from gprocess";
	cout << endl;
	exit(0);
}

void set_main_debug_option_parameter(char* pptr, string option)
{
	if (option == "ganalyse")
	{
		if (!pptr)
		{
			cout << "\t\tgprocess{summary|1:basic|2:intermediate|3:detail|4}, integer: default 0=off, no parameters sets summary" << endl;
		}
		else
		{
			if (!strcmp(pptr,"summary") || !strcmp(pptr,"1") )
			{
				debug_control::DEBUG = debug_control::SUMMARY;
				debug << "set_main_debug_option_parameter: setting debug option debug_control::DEBUG = debug_control::SUMMARY" << endl;		
			}
			if (!strcmp(pptr,"basic") || !strcmp(pptr,"2") )
			{
				debug_control::DEBUG = debug_control::BASIC;
				debug << "set_main_debug_option_parameter: setting debug option debug_control::DEBUG = debug_control::BASIC" << endl;		
			}
			else if (!strcmp(pptr,"intermediate") || !strcmp(pptr,"3"))
			{
				debug_control::DEBUG = debug_control::INTERMEDIATE;
				debug << "set_main_debug_option_parameter: setting debug option debug_control::DEBUG = debug_control::INTERMEDIATE" << endl;		
			}
			else if (!strcmp(pptr,"detail") || !strcmp(pptr,"4"))
			{
				debug_control::DEBUG = debug_control::DETAIL;
				debug << "set_main_debug_option_parameter: setting debug option debug_control::DEBUG = debug_control::DETAIL" << endl;		
			}
			else if (!strcmp(pptr,"exhaustive") || !strcmp(pptr,"5"))
			{
				debug_control::DEBUG = debug_control::EXHAUSTIVE;
				debug << "set_main_debug_option_parameter: setting debug option debug_control::DEBUG = debug_control::EXHAUSTIVE" << endl;		
			}
		}
	}
}

void main_display_default_options()
{
	cout << "\t\tganalyse{summary}" << endl;
}

void set_programme_short_option(char* cptr)
{
	if (strchr(cptr, 'H') && strchr(cptr, '!'))
	{
		if (strchr(cptr,'#'))
		{
			debug_help();
		}
		else
		{					
			helpinfo();
		}
	}

	char* dptr = strchr(cptr, '#');
	if (dptr)
	{
		debug.open ("ganalyse.dbg"); 

		if (!debug)
		{
			cout << "\nError opening debug file\n";
			exit(0);
		}
		else
		{
			debug << "Debug information from ganalyse\n\n";
			debug << boolalpha;
		}

		if (!debug_setup(cptr))
		{
			debug_control::DEBUG = debug_control::SUMMARY;
			debug << "ganalyse: default debug options set" << endl;
		}
		
	}

	if (strchr(cptr, 'c'))
	{
		DISTINCT_NON_TRIVIAL_CLASSES = true;
	}

	if (strchr(cptr, 'e'))
	{
		EQUIVALENCE_CLASS_DETAIL = true;
	}

	if (strchr(cptr, 'h'))
	{
		helpinfo();
		exit(0);
	}

	if (strchr(cptr, 'u'))
		UNORIENTED_SEARCH = true;

	if (strchr(cptr, 'X'))
		DEVELOPMENT_MODE = true;

}

void set_programme_long_option(char* cptr, string source)
{
if (debug_control::DEBUG >= debug_control::DETAIL)
	debug << "set_programme_long_option: passed " << cptr << " from " << source << endl;

	char  loc_buf[40];
	char* c1;
	char* c2;

	c1 = cptr;

	/* take out any leading -- */
	while (*c1 == '-')
		c1++;
	
	c2 = loc_buf;
	while (isalpha(*c1) || isdigit(*c1) || *c1 == '-' || *c1 == '_')
		*c2++ = *c1++;
    *c2 = '\0';
	
	if (!strcmp(loc_buf,"compare"))
	{
		COMPARE_CODES = true;
		
if (debug_control::DEBUG >= debug_control::SUMMARY)
	debug << "set_programme_long_option: read COMPARE_CODES from " << source << endl;
	}

	if (!strcmp(loc_buf,"moves"))
	{
		REIDEMEISTER_MOVES = true;
		
if (debug_control::DEBUG >= debug_control::SUMMARY)
	debug << "set_programme_long_option: read REIDEMEISTER_MOVES from " << source << endl;
	}
}

void print_prog_params (ostream& os, string level, string prefix){}

/* describe_reidemeister_moves displays the a sequence of Reidemeister moves that demonstrate the equivalence of the two Gauss codes corresponding to
   the code_list_elements at index_1 and index_2 of code_element, or, if index_2 == -1, between the Gauss code corresponding to the code_list_element
   at index_1 and the trivial (or disconnected) min_representation of the equivalence class to which it belongs.
   
   If index_2 != -1 it is the responsibility of the calling code to ensure that both the code_list_elements belong to the same equivalence class.  
   
   The function determines an optimal maximal tree within the graph corresponding to the equivalence class containing the two codes with root the first 
   code.  The Reidemeister_tree_node objects contain an index into code_element for the code_list_element corresponding to the tree node.  Since there 
   is no code_list_element for the unlink, the Reidemeister_tree_node object for the unlink has its index set to -1.

   The Reidemeister_tree_node objects are stored in a vector and we maintain a list<int>, node_list, of vector offsets to record the order in which the 
   Reidemeister_tree_node objects should be considered when extending the tree.  The index field in the Reidemeister_tree_node gives us the code_list_element
   containing the list_target (i.e. tree edges) to consider from each Reidemeister_tree_node.     Note that the indices in the code_list_elements belonging 
   to the equivalence class are not contiguous.
   
   We do not push onto the node_list the offset of a Reidemeister_tree_node corresponding to the unlink, since there is no code_list_element for the unlink and 
   therefore no list_targets (edges) to consider from that node.  In the case that we have a tree root (determined by index_1) containing a single list_target 
   to the unlink, then there are no other code_list_elements in the same class (otherwise the root node would have (at least a reverse) list target to another node.
   
   It then reports the Reidemeister moves required to manipulate the second code into the second code.  That is the set of moves in the path from the first code 
   up to the root of the tree.
   
*/

void describe_reidemeister_moves(vector<code_list_element>& code_element,vector<int>& class_element_count, int index_1, int index_2, string special_case_description)
{
	int class_identifier = code_element[index_1].class_identifier;
	int num_code_list_elements_in_class = class_element_count[class_identifier];
	
	/* we create a tree node for each code_list_element in the class that will record the parent node in the tree, the root_hop_count, 
	   the list_target that describes the Reidemeister moves between the node and its parent and total number of Reidemeister moves
	   between the code_list_element and the root.  We include space for a Reidemeister_tree_node corresponding to the unlink, noting 
	   that there is no code_list_element for the unlink.
	*/
	vector<Reidemeister_tree_node> tree_node(num_code_list_elements_in_class+1);			
	
	/* the first Reidemeister_tree_node in the vector corresponds to the root code_list_element, since it is the root of the tree, we need only update the index 
	   of the Reidemeister_tree_node.  Using root_index allows us to interchange the roles of index_1 and index_2 if required in the future.
	   
	   The integer variable node is used to count the number of nodes added to the tree and as an index into tree_node for the next Reidemeister_tree_node
	*/
	int root_index = index_1;	
	int leaf_index = index_2;
	int leaf_tree_node = -1;  // will record the index of the Reidemeinster_tree_node corresponding to the leaf index.		
	tree_node[0].index = root_index;
	int node=1; 

if (debug_control::DEBUG >= debug_control::SUMMARY)
{
	debug << "describe_reidemeister_moves: evaluating Reidemeister moves between root_index " << root_index << " and leaf index " << leaf_index << endl;
	debug << "describe_reidemeister_moves: create maximal tree through equivalence class with root " << root_index << endl;
}			
	/* node_list records the progress of building the tree, it stores the offset into tree_node of nodes to be processed in a breadth first search */
	list<int> node_list; 
	node_list.push_back(0);

	while (node_list.size() > 0)
	{				
		int next_tree_node_offset = *(node_list.begin());

		Reidemeister_tree_node& next_tree_node = tree_node[next_tree_node_offset];			
		code_list_element& next_list_element = code_element[tree_node[next_tree_node_offset].index]; // we don't push the unlink tree node onto the list, so index is always valid
		node_list.pop_front();

if (debug_control::DEBUG >= debug_control::SUMMARY)
{
	debug << "describe_reidemeister_moves: next_tree_node_offset  = " << next_tree_node_offset << " tree node: " << endl;
	print_Reidemeister_tree_node(debug, next_tree_node, "describe_reidemeister_moves:   ");
	debug << "describe_reidemeister_moves: next_list_element.index = " << next_list_element.index << endl;
}
				
if (debug_control::DEBUG >= debug_control::SUMMARY)
	debug << "describe_reidemeister_moves: next_list_element.num_targets = " << next_list_element.num_targets << endl;
				
		for (int i=0; i< next_list_element.num_targets; i++)
		{
			int neighbour_tree_node_index = tree_node_index(tree_node,node,next_list_element.target[i].location);

if (debug_control::DEBUG >= debug_control::SUMMARY)
	debug << "describe_reidemeister_moves:   list_target " << i << " points to neighbour_tree_node_index " << neighbour_tree_node_index << endl;
					
			if (neighbour_tree_node_index == -1) // includes the case where location is -1, indicating the unlink and there is no existing tree_node for the unlink
			{	
if (debug_control::DEBUG >= debug_control::SUMMARY)
	debug << "describe_reidemeister_moves:     no corresponding Reidemeister_tree_node" << endl;

				/* There is no Reidemeister_tree_node corresponding to the neighbour code_list_element 
				   so create a new Reidemeister_tree_node for next_list_element.target[i].location 
				*/
				tree_node[node].index = next_list_element.target[i].location;
				tree_node[node].parent = next_list_element.index;
				tree_node[node].parent_tree_node_offset = next_tree_node_offset;
				tree_node[node].root_hop_count = next_tree_node.root_hop_count + 1;					

				if (tree_node[node].index == -1) 
				{
					/* list_target points to the unlink, so we record the 'downward' list_target, from parent to node, rather than the' upward'
					   list target from the node to the parent 
					*/
					tree_node[node].root_list_target = i;
					tree_node[node].parent_list_target = true;
					tree_node[node].num_R1_moves = next_tree_node.num_R1_moves + next_list_element.target[i].R1_move.size();
					tree_node[node].num_R2_moves = next_tree_node.num_R2_moves + next_list_element.target[i].R2_move.numrows();
					tree_node[node].num_R3_moves = next_tree_node.num_R3_moves + next_list_element.target[i].R3_move.numrows();							
				}
				else
				{
					/* There is a code_list_element corresponding to the neighbour.  Identify the list target at next_list_element.target[i].location 
					   that points back to next_list_element 
					*/
					code_list_element& neighbour_list_element = code_element[next_list_element.target[i].location];
					int reverse_list_target_index_at_neighbour = -1;
					for (int j=0; j< neighbour_list_element.num_targets; j++)
					{
						if(neighbour_list_element.target[j].location == next_list_element.index)
						{
							reverse_list_target_index_at_neighbour = j;
							break;
						}
					}

					list_target& neighbour_reverse_list_target = neighbour_list_element.target[reverse_list_target_index_at_neighbour];
					
					int num_moves = neighbour_reverse_list_target.R1_move.size()+neighbour_reverse_list_target.R2_move.numrows()+neighbour_reverse_list_target.R3_move.numrows();
					
					if (num_moves == 0)
					{
						/* the reverse list target was added as a result of a Reidemeister II move and has no moves associated with it, 
						   so we record the 'downward' list_target, from parent to node, rather than the' upward' list target from the node 
						   to the parent 
						*/
						tree_node[node].root_list_target = i;
						tree_node[node].parent_list_target = true;
						tree_node[node].num_R1_moves = next_tree_node.num_R1_moves + next_list_element.target[i].R1_move.size();
						tree_node[node].num_R2_moves = next_tree_node.num_R2_moves + next_list_element.target[i].R2_move.numrows();
						tree_node[node].num_R3_moves = next_tree_node.num_R3_moves + next_list_element.target[i].R3_move.numrows();
					}
					else
					{
						tree_node[node].root_list_target = reverse_list_target_index_at_neighbour;
						tree_node[node].num_R1_moves = next_tree_node.num_R1_moves + neighbour_reverse_list_target.R1_move.size();
						tree_node[node].num_R2_moves = next_tree_node.num_R1_moves + neighbour_reverse_list_target.R2_move.numrows();
						tree_node[node].num_R3_moves = next_tree_node.num_R1_moves + neighbour_reverse_list_target.R3_move.numrows();
					}							
				}
						
if (debug_control::DEBUG >= debug_control::SUMMARY)
{
	debug << "describe_reidemeister_moves:     created tree_node index " << node << " corresponding to code_list_element " << tree_node[node].index << ", code ";
	if (tree_node[node].index == -1) 
		debug << "unlink" << endl;
	else
	{
		code_list_element& neighbour_list_element = code_element[next_list_element.target[i].location];
		debug << neighbour_list_element.code << endl;	
	}
	debug << "describe_reidemeister_moves:     tree_node has root_hop_count = " << tree_node[node].root_hop_count << ", " << tree_node[node].num_R1_moves << " RI moves, " << tree_node[node].num_R2_moves << " R2 moves, " << tree_node[node].num_R3_moves << " R3 moves" << endl;	
}
				if (tree_node[node].index == leaf_index)
				{
					leaf_tree_node = node;
if (debug_control::DEBUG >= debug_control::SUMMARY)
	debug << "describe_reidemeister_moves: encountered leaf_index " << leaf_index << ", leaf_tree_node = " << leaf_tree_node << endl;
				}

				if (tree_node[node].index != -1)
					node_list.push_back(node); 
					
				node++;
			}
			else // includes the case where location is -1, indicating the unlink and there is already a tree_node for the unlink
			{
				/* check the neighbouring Reidemeister_tree_node to see if the next_list_element offers a better path to the root than the 
				   currently held by the neighbourng Reidemeister_tree_node 
				*/
				Reidemeister_tree_node& neighbour_tree_node = tree_node[neighbour_tree_node_index];
				int reverse_list_target_index_at_neighbour = -1;						
				bool preferred = false;

if (debug_control::DEBUG >= debug_control::SUMMARY)
{
	code_list_element& neighbour_list_element = code_element[next_list_element.target[i].location];
	debug << "describe_reidemeister_moves:   neighbour tree_node " << neighbour_tree_node_index << " corresponds to code_list_element " << neighbour_list_element.index << ", code " << neighbour_list_element.code <<  endl;
	debug << "describe_reidemeister_moves:   neighbour tree_node has root_hop_count = " << neighbour_tree_node.root_hop_count << ", " << neighbour_tree_node.num_R1_moves << " RI moves, " << neighbour_tree_node.num_R2_moves << " R2 moves, " << neighbour_tree_node.num_R3_moves << " R3 moves" << endl;	
}
						
				if (neighbour_tree_node.index == -1) 
				{
if (debug_control::DEBUG >= debug_control::SUMMARY)
{
	debug << "describe_reidemeister_moves:     downwards list target to neighbour:" <<  endl;
	print_list_target(debug,next_list_element.target[i],"describe_reidemeister_moves:     ");
}	
							
					preferred = preferred_parent(next_tree_node, next_list_element.target[i], neighbour_tree_node);
				}
				else
				{
					code_list_element& neighbour_list_element = code_element[neighbour_tree_node.index]; 

					/* Identify the list target at the neighbour code_list_element that points back to next_list_element */
					for (int j=0; j< neighbour_list_element.num_targets; j++)
					{
						if(neighbour_list_element.target[j].location == next_list_element.index)
						{
							reverse_list_target_index_at_neighbour = j;
							break;
						}
					}

					list_target& neighbour_reverse_list_target = neighbour_list_element.target[reverse_list_target_index_at_neighbour];

if (debug_control::DEBUG >= debug_control::SUMMARY)
{
	debug << "describe_reidemeister_moves:     upwards list target from neighbour:" <<  endl;
	print_list_target(debug,neighbour_reverse_list_target,"describe_reidemeister_moves:     ");
}	

					preferred = preferred_parent(next_tree_node, neighbour_reverse_list_target, neighbour_tree_node);							
				}
				
				if (preferred)
				{
if (debug_control::DEBUG >= debug_control::SUMMARY)
	debug << "describe_reidemeister_moves:   code list element " << next_list_element.index << " offers preferred path to root" <<  endl;

					neighbour_tree_node.parent = next_list_element.index;
					neighbour_tree_node.parent_tree_node_offset = next_tree_node_offset;
					neighbour_tree_node.root_hop_count = next_tree_node.root_hop_count+1;
					
					if (neighbour_tree_node.index == -1) 
					{
						neighbour_tree_node.root_list_target = i;  
						neighbour_tree_node.parent_list_target = true;
						neighbour_tree_node.num_R1_moves = next_tree_node.num_R1_moves + next_list_element.target[i].R1_move.size();
						neighbour_tree_node.num_R2_moves = next_tree_node.num_R2_moves + next_list_element.target[i].R2_move.numrows();
						neighbour_tree_node.num_R3_moves = next_tree_node.num_R3_moves + next_list_element.target[i].R3_move.numrows();								
					}
					else
					{
						code_list_element& neighbour_list_element = code_element[neighbour_tree_node.index]; 
						list_target& neighbour_reverse_list_target = neighbour_list_element.target[reverse_list_target_index_at_neighbour];								
						int num_moves = neighbour_reverse_list_target.R1_move.size()+neighbour_reverse_list_target.R2_move.numrows()+neighbour_reverse_list_target.R3_move.numrows();
						
						if (num_moves == 0)
						{
							neighbour_tree_node.root_list_target = i;  
							neighbour_tree_node.parent_list_target = true;
							neighbour_tree_node.num_R1_moves = next_tree_node.num_R1_moves + next_list_element.target[i].R1_move.size();
							neighbour_tree_node.num_R2_moves = next_tree_node.num_R2_moves + next_list_element.target[i].R2_move.numrows();
							neighbour_tree_node.num_R3_moves = next_tree_node.num_R3_moves + next_list_element.target[i].R3_move.numrows();
						}
						else
						{
							neighbour_tree_node.root_list_target = reverse_list_target_index_at_neighbour;
							neighbour_tree_node.num_R1_moves = next_tree_node.num_R1_moves + neighbour_reverse_list_target.R1_move.size();
							neighbour_tree_node.num_R2_moves = next_tree_node.num_R1_moves + neighbour_reverse_list_target.R2_move.numrows();
							neighbour_tree_node.num_R3_moves = next_tree_node.num_R1_moves + neighbour_reverse_list_target.R3_move.numrows();
						}
					}
				}
				else
				{
if (debug_control::DEBUG >= debug_control::SUMMARY)
	debug << "describe_reidemeister_moves:   code list element " << next_list_element.index << " does not offer a preferred path to root" <<  endl;
				}
			}
		}
	}
			
if (debug_control::DEBUG >= debug_control::SUMMARY)
	debug << "describe_reidemeister_moves: created  = " << node << " Reidemeister_tree_node objects" << endl;
			
			
	/* process the moves in the maximal tree we have created */
	int tree_node_offset = leaf_tree_node;
	do
	{
		Reidemeister_tree_node& next_tree_node = tree_node[tree_node_offset];

if (debug_control::DEBUG >= debug_control::SUMMARY)
	debug << "describe_reidemeister_moves: next_tree_node.index = " << next_tree_node.index << ", parent = " << next_tree_node.parent << ", root_list_target = " << next_tree_node.root_list_target << endl;

		if(next_tree_node.parent_list_target)
		{	
			cout << "  Diagram " << (next_tree_node.index == -1? special_case_description: code_element[next_tree_node.index].code) << " may be obtained from " << code_element[next_tree_node.parent].code << " by:" << endl;
			print_Reidemeister_moves(cout,code_element[next_tree_node.parent].target[next_tree_node.root_list_target]);

			output << "    Diagram " << (next_tree_node.index == -1? special_case_description: code_element[next_tree_node.index].code) << " may be obtained from " << code_element[next_tree_node.parent].code << " by:" << endl;
			print_Reidemeister_moves(output,code_element[next_tree_node.parent].target[next_tree_node.root_list_target]);
		}
		else
		{
			cout << "  Diagram " << code_element[next_tree_node.parent].code << " may be obtained from " << (next_tree_node.index == -1? special_case_description: code_element[next_tree_node.index].code) << " by:" << endl;
			print_Reidemeister_moves(cout,code_element[next_tree_node.index].target[next_tree_node.root_list_target]);

			output << "    Diagram " << code_element[next_tree_node.parent].code << " may be obtained from " << (next_tree_node.index == -1? special_case_description: code_element[next_tree_node.index].code) << " by:" << endl;
			print_Reidemeister_moves(output,code_element[next_tree_node.index].target[next_tree_node.root_list_target]);
		}
		
		tree_node_offset = next_tree_node.parent_tree_node_offset;

	} while (tree_node_offset != 0);	
}
