/********************************************************************************************************************
glist enumerates Gauss codes for a given number of components and number of crossings.  It is based on Jeremy Green's
approach of building unsigned Gauss codes recursively by numbering crossings in the order in which they are first 
encountered.  Reidemeister I configurations and cyclic rotations are removed in a similar manner to Jeremy's code.

Crossing signs are then appended to the unsigned codes  so that glist produces the same list of codes for one 
component as Jeremy's genallgauss programme does.  Note, however, that this approach does not produce an optimal 
list, in the sense that some of the codes have the same unoriented over preferred Gauss code but it is considerably 
quicker to calcuate the over-preferred code and remove duplicates from the list produced by glist than to do so as
we go along.

For example, the two Gauss codes O1-O2+O3+U1-U2+U3+ and O1+O2+O3-U1+U2+U3- are produced using this algorithm, they both 
represent the unknot and have the same unoriented over-preferred Gauss code O1+O2+O3-U1+U2+U3-

Note also that even without signs, the minimal cyclic representation of a Gauss code, numbering crossings in the order
in which they are encountered is not the same as an over-preferred representation.  For example, the unsigned Gauss code

O1 U2 O3 U4 O5 U1 O6 U3 O2 U5 O4 U6

is cyclically minimal but is not over-preferred in the sense that O1 O2 O3 O4 O5 O6 is not a subsequence of this code.

We use the same encoding for Gauss code terms as Jeremy does; that is, each term is an integer of the form
6*<crossing_number>+<crossing_detail>, where <crossing_detail>=enum{O-, O, O+, U-, U, U+}

**********************************************************************************************************************/
using namespace std;
#include <fstream>
#include <iostream>
#include <sstream>
#include <cstdlib>
#include <cstring>
#include <vector>
#include <list>
#include <algorithm>
#include <csignal>
#include <iomanip>
#include <ctype.h>

/******************** Global Variables **********************/
bool			COMPONENT_BOUNDARIES = false;
bool 			UNORIENTED_SEARCH = false;
bool 			SIGNED_OUTPUT = false;
bool 			WRITE_OUTPUT = true;

bool			sigusr1_request = false;
bool			termination_request = false;
int				num_crossings=0;
int				num_code_terms=0;
int				num_components=0;
int 			shift_power = 23;  // the hashed code list vector is 2^shift_power long.
bool     		DEVELOPMENT_MODE = false;

ofstream 	output;
extern ofstream    debug;

#include <glist.h>

/********************* Function prototypes ***********************/
void check_debug_option_parameters(char* pptr, string option);
bool debug_setup(char* argv_ptr);
void debug_help();
bool next_label(vector<int>& array, int& num_classifying_crossings);
void helpinfo();
void set_programme_long_option(char* cptr, string source);
void set_programme_short_option(char* cptr);
void print_prog_params (ostream& os, string level, string prefix);

void generate_codes(list<Gauss_code>& unsigned_codes, vector<code_term_type>& term_values, vector<bool>& used, Gauss_code& candidate_code, vector<int>& component_map, int max, int num_terms, int level);
vector<code_term_type> minimal_cyclic_code(Gauss_code& unsigned_code);
bool Reidemeister_I_present(Gauss_code& c);


namespace util
{
	string itos(long n);
}

void sigterm_handler (int sig) // kill -15, or just kill
{
	cout << "\nglist programme received SIGTERM, setting termination_request=true" << endl;
	termination_request = true;
}

/* sigusr1_handler causes glist to report progress on the current task and continue. */
void sigusr1_handler (int sig) // kill -10
{
	cout << "\nglist programme received SIGUSR1, setting sigusr1_request=true" << endl;
	sigusr1_request = true;
}

/******************* Main Function ************************/

int main (int argc, char* argv[])
{	
	signal(SIGTERM, sigterm_handler); // register the SIGTERM handler
	signal(SIGUSR1, sigusr1_handler); // register the SIGUSR1 handler

	
	/* take a copy of the original command line in case we're debugging and remove
	   the debug details.  We will want to record the original command line for the debug file.
	*/
	
	vector<string> original_programme_arguments(argc); 
	for (int i = 0; i< argc; i++)
		original_programme_arguments[i] = string(argv[i]);
	
	ifstream    input;

	ofstream 		tempfile;

	time_t start_time = time(0);
	std::tm* start_timeinfo = localtime (&start_time);
	
    /* Read the programme parameters and input files */
	for (int i=1; i < argc; i++)
	{
		if (*(argv[i]) == '-')
		{
			if (*(argv[i]+1) == '-')
				set_programme_long_option(argv[i], "command line");
			else 
				set_programme_short_option(argv[i]);
		}
	}	
	
if (debug_control::DEBUG >= debug_control::SUMMARY)
	print_prog_params (debug, "basic", "");

	if (!DEVELOPMENT_MODE && (!num_crossings || !num_components ))
	{
		cout << "\n";
		if (!num_crossings)
			cout << "num-crosings option required" << endl;
			
		if (!num_components)
			cout << "num-cpts option required" << endl;
			
		helpinfo();
		exit(0);
	}

if (debug_control::DEBUG >= debug_control::SUMMARY)
{
	debug << "full command line: ";
	for (int i = 0; i< argc; i++)
		debug << original_programme_arguments[i] << ' ';
	debug << endl;
}


/*DEVELOPMENT_MODE is used for testing individual functions during development */
if (DEVELOPMENT_MODE)
{
	exit(0);
	/**************/
}
/* End of DEVELOPMENT_MODE code */
		
if (debug_control::DEBUG >= debug_control::SUMMARY)
{
	debug << "main: num_crossings = " << num_crossings << endl;
	debug << "main: num_code_terms = " << num_code_terms << endl;
	debug << "main: num_components = " << num_components << endl;
}
	
    output.open ("glist.out");

    if (!output)
    {
		cout << "\nError opening output file\n";
		exit(0);
	}
	else
	{
		output << "Output from glist ";
		for (int i=1; i< argc; i++	)
			output << argv[i] << " ";
		output << "\n";
		
		output << "Start time: " << asctime(start_timeinfo);
	}

if (debug_control::DEBUG >= debug_control::DETAIL)
	debug << "gauss: enumerating num_crossings = " << num_crossings << endl;

	list<Gauss_code> unsigned_codes;
	
	vector<code_term_type> term_values(2*num_crossings);
	for (int i=0; i< 2 * num_crossings; i++)
		term_values[i] = 3*i+1;  // encodes unsigned crossings O1, U1, O2, U2,...

	Gauss_code candidate_code(2*num_crossings, num_components);
	

	/* If we are working with multiple components we consider every combination of component boundaries within the Gauss code.  These can occur after 
	   indices 0,...,num_code_terms-2 in a vector of Gauss code elements.
	   
	   Each top-level call to generate codes produces "minimal" Gauss codes with the same, fixed, component boundaries.  We do not permute components
	   when looking for a minimal code.  This has the effect of increasing the size of the list that we produce, since two codes may be equivalent if
	   we consider the components in a different order.  However, this approach makes the search for minimal codes much faster, and we will be filtering
	   the list for over-preferred Gauss codes once the list has been produced, so any duplicates will be removed at that stage.
	*/
	vector<int> component_boundary(num_components-1);
	if (num_components > 1)
	{
		for (int i=0; i< num_components-1; i++)
			component_boundary[i] = i;
	}
	
	bool another_component_combination; // set at the end of the loop clause
	do
	{						

if (num_components > 1 && debug_control::DEBUG >= debug_control::SUMMARY)
{
	debug << "component boundaries: ";
	for (int i=0; i<num_components-1; i++)
		debug << component_boundary[i] << " ";
	debug << endl;	
}

		vector<int> num_terms_in_component(num_components);
		num_terms_in_component[0] = (num_components == 1? num_code_terms: component_boundary[0]+1);
		
		for (int i=1; i< num_components-1; i++)
			num_terms_in_component[i] = component_boundary[i] - component_boundary[i-1];

		if (num_components > 1)
			num_terms_in_component[num_components-1] = num_code_terms - component_boundary[num_components-2]-1;
			

if (num_components > 1 && debug_control::DEBUG >= debug_control::SUMMARY)
{
	debug << "num_terms_in_component: ";
	for (int i=0; i<num_components; i++)
		debug << num_terms_in_component[i] << " ";
	debug << endl;	
}

		vector<bool> used(num_code_terms);
	
		/* component map[i] = c when the ith code term should be considered to belong to component c, counting from 1 */
		vector<int> component_map(num_code_terms);
		if (num_components == 1)
		{
			for(int i=0; i< num_code_terms; i++)
				component_map[i] = 1;
		}
		else
		{
			int index = 0;
			int component = 1;
			for (int i=0; i< num_components; i++)
			{
				for (int j=0; j<num_terms_in_component[i]; j++)
					component_map[index++] = component;

				component++;
			}
		}

if (num_components > 1 && debug_control::DEBUG >= debug_control::SUMMARY)
{
	debug << "component_map: ";
	for (int i=0; i<num_code_terms; i++)
		debug << component_map[i] << " ";
	debug << endl;	
}
		
		candidate_code.num_terms_in_component = num_terms_in_component;

		/* Note that due to the way that minimal_cyclic code operates, generate_codes always produces codes starting with O1, regardless of the number of components */
		if (!COMPONENT_BOUNDARIES)
			generate_codes(unsigned_codes, term_values, used, candidate_code, component_map, 2, 0, 0); // max=2, num_terms=0, level = 0
		
		/* look for another component combination, we can place component markers after indices 0,...,num_code_terms-2 */
		another_component_combination = false;
		if (num_components > 1)
		{
			for (int i=num_components-2; i>=0; i--)
			{
				if (component_boundary[i] < (num_code_terms - num_components + i))
				{
					/* increment the ith boundary marker and reset all subsequet markers */
					component_boundary[i]++;
					for (int j=i+1; j< num_components-1; j++)
						component_boundary[j] = component_boundary[i] + j - i;
					another_component_combination = true;
					break;
				}
			}
		}		
	} while (another_component_combination);


	
	output << "generated " << unsigned_codes.size() << " codes" << endl;
	list<Gauss_code>::iterator lptr = unsigned_codes.begin();
	
	while (lptr != unsigned_codes.end())
	{
		if (SIGNED_OUTPUT)
		{
			write_signed_code(output, *lptr);
			output << endl;
		}
		else if (WRITE_OUTPUT)
		{
			write_gauss_code(output, *lptr);
			output << endl;
		}
		else
		{
			print_gauss_code(output, *lptr);
			output << endl;
		}
		
		lptr++;
	}
	

	time_t end_time = time(0);
	double time_exp = difftime(end_time, start_time);
	int dd = static_cast<int>(time_exp/3600/24);
	time_exp -= dd*3600*24;
	int hh = static_cast<int>(time_exp/3600);
	time_exp -= hh*3600;
	int mm = static_cast<int>(time_exp/60);
	time_exp -= mm*60;
	int ss = static_cast<int>(time_exp);
	output << "\nRunning time: " << dd << " days " << hh << " hours " << mm	<< " minutes " << ss << " seconds" << endl;

	output.close();
	cout << endl;

	return 0;
} /* End of main */


void helpinfo()
{
	cout << "\nUsage: glist [-<short-option>] [--<long-option>] ";
	cout << "\n<short-option>:";
	cout << "\n       #: debug, see #H! for details";
	cout << "\n       b: show component boundaries only (requires debug)";
	cout << "\n       h: show help";
	cout << "\n       u: unoriented search";
	cout << "\n       S: produce signed Gauss code output";
	cout << "\n       W: produce standard Gauss code output rather than OU format";
	cout << "\n       #H!: display debug help screen";
	cout << "\n<long-option>:";
	cout << "\n       num-cpts=n: number of components";
	cout << "\n       num-crossings=n: number of crossings in the Gauss codes";
	cout << endl;
	exit(0);
}

/***************  Functions required for setting up glist specific debug **************/

void set_main_debug_option_parameter(char* pptr, string option);
void set_main_debug_option(char* start, char* end)
{
	char  loc_buf[end-start+2];
	char* c1 = start;
	char* c2 = loc_buf;

	/* if both start and end are zero, display debug help information.
	   this has been included here in this manner so that each time a debug option
	   is added, the help will be updated (hopefully!)
	*/
	if (start == 0 && end == 0)
	{
		set_main_debug_option_parameter(0,"glist");
		return;
	}

	do
	{
		*c2++ = *c1++;
	} while (c1 <= end);
	
	*c2 = '\0';

	char* pptr = strchr(loc_buf,'{');
	if (pptr)
	{
		*pptr++ = '\0';
	}

	/* now, even if there are parameters, the first part of loc_buf 
	   is a C-string that identifies the option */
	
	if (!strcmp(loc_buf,"glist"))
	{
		debug_control::DEBUG = debug_control::SUMMARY;
		debug << "main::set_main_debug_option: setting debug option debug_control::DEBUG = debug_control::SUMMARY" << endl;		
		
		if (pptr)
		{
			/* check for any parameters */
			check_debug_option_parameters(pptr, "glist");
		}
	}

	debug.flush();

}

void set_main_debug_option_parameter(char* pptr, string option)
{
	if (option == "glist")
	{
		if (!pptr)
		{
			cout << "\t\tglist{summary|1:basic|2:intermediate|3:detail|4}, integer: default 0=off, no parameters sets summary" << endl;
		}
		else
		{
			if (!strcmp(pptr,"summary") || !strcmp(pptr,"1") )
			{
				debug_control::DEBUG = debug_control::SUMMARY;
				debug << "main::set_main_debug_option_parameter: setting debug option debug_control::DEBUG = debug_control::SUMMARY" << endl;		
			}
			if (!strcmp(pptr,"basic") || !strcmp(pptr,"2") )
			{
				debug_control::DEBUG = debug_control::BASIC;
				debug << "main::set_main_debug_option_parameter: setting debug option debug_control::DEBUG = debug_control::BASIC" << endl;		
			}
			else if (!strcmp(pptr,"intermediate") || !strcmp(pptr,"3"))
			{
				debug_control::DEBUG = debug_control::INTERMEDIATE;
				debug << "main::set_main_debug_option_parameter: setting debug option debug_control::DEBUG = debug_control::INTERMEDIATE" << endl;		
			}
			else if (!strcmp(pptr,"detail") || !strcmp(pptr,"4"))
			{
				debug_control::DEBUG = debug_control::DETAIL;
				debug << "main::set_main_debug_option_parameter: setting debug option debug_control::DEBUG = debug_control::DETAIL" << endl;		
			}
			else if (!strcmp(pptr,"exhaustive") || !strcmp(pptr,"5"))
			{
				debug_control::DEBUG = debug_control::EXHAUSTIVE;
				debug << "main::set_main_debug_option_parameter: setting debug option debug_control::DEBUG = debug_control::EXHAUSTIVE" << endl;		
			}
		}
	}
}

void main_display_default_options()
{
	cout << "\t\tglist{summary}" << endl;
}

void set_programme_short_option(char* cptr)
{
	if (strchr(cptr, 'H') && strchr(cptr, '!'))
	{
		if (strchr(cptr,'#'))
		{
			debug_help();
		}
		else
		{					
			helpinfo();
		}
	}

	char* dptr = strchr(cptr, '#');
	if (dptr)
	{
		debug.open ("glist.dbg"); 

		if (!debug)
		{
			cout << "\nError opening debug file\n";
			exit(0);
		}
		else
		{
			debug << "Debug information from glist\n\n";
			debug << boolalpha;
		}

		if (!debug_setup(cptr))
		{
			debug_control::DEBUG = debug_control::SUMMARY;
			debug << "main: default debug options set" << endl;
		}
		
	}


	if (strchr(cptr, 'b'))
		COMPONENT_BOUNDARIES = true;

	if (strchr(cptr, 'h'))
		helpinfo();

	if (strchr(cptr, 'u'))
		UNORIENTED_SEARCH = true;

	if (strchr(cptr, 'S'))
		SIGNED_OUTPUT = true;

	if (strchr(cptr, 'W'))
		WRITE_OUTPUT = false;

	if (strchr(cptr, 'X'))
		DEVELOPMENT_MODE = true;
}

void set_programme_long_option(char* cptr, string source)
{
if (debug_control::DEBUG >= debug_control::DETAIL)
	debug << "set_programme_long_option: passed " << cptr << " from " << source << endl;

	char  loc_buf[40];
	char* c1;
	char* c2;

	c1 = cptr;

	/* take out any leading -- */
	while (*c1 == '-')
		c1++;
	
	c2 = loc_buf;
	while (isalpha(*c1) || isdigit(*c1) || *c1 == '-' || *c1 == '_')
		*c2++ = *c1++;
    *c2 = '\0';
	
	if (!strcmp(loc_buf,"num-cpts"))
	{
		if (*c1 == '=')
		{
			get_number(num_components,++c1);
		}
		else
		{
			cout << "\nYou must specify the number of components with the num-cpts option, e.g. num-cpts=2" << endl;
			exit(0);
		}
		
if (debug_control::DEBUG >= debug_control::SUMMARY)
	debug << "set_programme_long_option: num-cpts " << source << ", num-cpts = " << num_components << endl;
	}
	else if (!strcmp(loc_buf,"num-crossings"))
	{
		if (*c1 == '=')
		{
			get_number(num_crossings,++c1);
			num_code_terms = 2* num_crossings;
		}
		else
		{
			cout << "\nYou must specify the number of crossings with the num-crossings option, e.g. num-crossings=7" << endl;
			exit(0);
		}
		
if (debug_control::DEBUG >= debug_control::SUMMARY)
	debug << "set_programme_long_option: num-crossings " << source << ", num-crossings = " << num_crossings << endl;
	}
	else
	{
if (debug_control::DEBUG >= debug_control::SUMMARY)
	debug << "set_programme_long_option: invalid long option " << loc_buf << " read from " << source << endl;
	
		cout << "Invalid long option " << loc_buf << endl;
		exit(0);		
	}
}

void print_prog_params (ostream& os, string level, string prefix)
{
    os << prefix << "num_components = " << num_components << endl;
    os << prefix << "num_crossings = " << num_crossings << endl;
}

		
void generate_codes(list<Gauss_code>& unsigned_codes, vector<code_term_type>& term_values, vector<bool>& used, Gauss_code& candidate_code, vector<int>& component_map, int max, int num_terms, int level)
{	
	
if (debug_control::DEBUG >= debug_control::SUMMARY)
	debug << "generate_codes(" << level << "): entered with num_terms = " << num_terms << ", num_code_terms = " << num_code_terms << ", max = " << max << endl;


	if (sigusr1_request)
	{
		/* report progress */
		cout << "generate_codes: list size = " << unsigned_codes.size() << " candidate_code ";
		write_gauss_code(cout, candidate_code, num_terms);		
		sigusr1_request = false;
	}

	if (num_terms == num_code_terms) // code is complete and ready to add
	{					

if (debug_control::DEBUG >= debug_control::SUMMARY)
{
	debug << "generate_codes(" << level << "): presented with candidate_code ";
	write_gauss_code(debug,candidate_code);
}	             
	    vector<code_term_type> minimal_code = minimal_cyclic_code(candidate_code);
	    
		if (minimal_code != candidate_code.code)
		{       
if (debug_control::DEBUG >= debug_control::SUMMARY)
{
	debug << "generate_codes(" << level << "): rejected due to ";
	for (int i=0; i< num_code_terms; i++)
		debug << minimal_code[i] << ' ';
	debug << endl;
}	  			
			return;
		}
		
		/* check that the code describes a connected diagram; a unicursal component disconnected iff it includes 
		   both the over and under crossings of all the crossings it encounters 
		*/
		if (num_components > 1)
		{
			int index = 0;
			bool connected = true;
			
			for (int i=0; i< num_components; i++)
			{
				connected = false;
				vector<int> crossing_count(num_crossings);
				for (int j=0; j< candidate_code.num_terms_in_component[i]; j++)
				{
					crossing_count[candidate_code.code[index]/6]++;
					index++;
				}
					
				for(int j=0; j< num_crossings; j++)
				{
					if (crossing_count[j]%2 == 1)
					{
						connected = true;
						break;
					}
				}
				
				if (!connected)
					break;
			}
			
			if (!connected)
				return;
		}

if (debug_control::DEBUG >= debug_control::SUMMARY)
{
	debug << "generate_codes(" << level << "): accepted ";
	for (int i=0; i< num_code_terms; i++)
		debug << minimal_code[i] << ' ';
	debug << endl;
}

		unsigned_codes.push_back(candidate_code);
	}
	else
	{
if (debug_control::DEBUG >= debug_control::SUMMARY)
{
	debug << "generate_codes(" << level << "): presented with candidate_code ";
	write_gauss_code(debug,candidate_code,num_terms);	
	debug << "generate_codes(" << level << "): max position = " << max << ", used: ";
	for (int i=0; i< num_code_terms; i++)
		debug << (i%2? 'U':'O') << i/2 + 1 << ':' << used[i] << ' ';
	debug << endl;
}	  
		for (int i = 0; i < max; i++) 
		{ // try different choices for next element

if (debug_control::DEBUG >= debug_control::SUMMARY)
	debug << "generate_codes(" << level << "):   term " << i << ", candidate value " << term_values[i] << ", " << (term_values[i]%6 == 1? 'O' : 'U') << term_values[i]/6 + 1;

			if (!used[i]) 
			{
				/* We don't want to use the next term if it creates a Reidemeister I configuration.  That will be the case if the crossing number of the next term is the same as the last term
				   and they are in the same component, or if the next term is the last term of the component and the crossing number is the same as the first term of the component.
				   
				   We also don't want to create codes whose first component does not contain an over-crossing, since these will never be minimal once we permute components.
				*/
if (debug_control::DEBUG >= debug_control::SUMMARY)
	debug << " unused" << endl;
	
				bool valid_next_term = true;
				if (num_terms != 0 )
				{
					/* if component_map[num_terms] != component_map[num_terms-1], we're choosing a value for the first term of a component */
					if (component_map[num_terms] == component_map[num_terms-1])
					{
						if (term_values[i]/6 == candidate_code.code[num_terms-1]/6) 
						{
if (debug_control::DEBUG >= debug_control::SUMMARY)
	debug << "generate_codes(" << level << "):     candidate value has same crossing as previous, creates R1 " << endl;
							valid_next_term = false;
						}
						else if (num_terms == num_code_terms-1 || component_map[num_terms] != component_map[num_terms+1]) 
						{
if (debug_control::DEBUG >= debug_control::SUMMARY)
	debug << "generate_codes(" << level << "):     reached the end of component " << component_map[num_terms] << endl;
	
							/* we're at the end of a component containing at least two terms, so we need to find the start of the component and check we're not creating a 
							   cyclic Reidemeister I.  If this is the first component, as we look back over the component, note whether it contains an over-crossing to make sure we're
							   not creating a first component comprise solely of under-crossings.
							*/
							bool over_crossing_found = false;
							
							int term = num_terms;
							do
							{
								term--;
								
					 			if (candidate_code.code[term]%6 == 1)
									over_crossing_found = true;
													 			
							} while (term > 0 && component_map[term-1] == component_map[num_terms]);

if (debug_control::DEBUG >= debug_control::SUMMARY)
{
	debug << "generate_codes(" << level << "):     first term at index " << term << ", value " << candidate_code.code[term]/6 << (term_values[i]%6 == 1? 'O' : 'U') << endl;
	debug << "generate_codes(" << level << "):     over-crossing found in component = " << over_crossing_found << endl;
}	

								
							if (term_values[i]/6 == candidate_code.code[term]/6 || (component_map[num_terms] == 1 && !over_crossing_found)) 
							{
if (debug_control::DEBUG >= debug_control::SUMMARY)
	debug << "generate_codes(" << level << "):     candidate value has same crossing as first term of component, or produces component with no over-crossings " << endl;
	
								valid_next_term = false;
							}						
						}
					}				
				}
				
				if (valid_next_term)
				{
if (debug_control::DEBUG >= debug_control::SUMMARY)
	debug << "generate_codes(" << level << "):      term is valid" << endl;
					used[i] = true;
					candidate_code.code[num_terms] = term_values[i];
					int new_max = (i >= max - 2 && max < num_code_terms) ? max + 2 : max;
					generate_codes(unsigned_codes, term_values, used, candidate_code, component_map, new_max, num_terms + 1, level+1);
					used[i] = false;
				}
				else
				{
if (debug_control::DEBUG >= debug_control::SUMMARY)
	debug << "generate_codes(" << level << "):      term is invalid" << endl;
				}
							
			}
			else
			{
if (debug_control::DEBUG >= debug_control::SUMMARY)
	debug << " used" << endl;
			}
		}
	}
}

