/**********************************************************************************************************************
gprocess takes as input a collection of files, "glist-i", produced by glist for i=2,...,max-crossings from a directory
identified by the filepath option, containing distinct Gauss codes for a given number of components.  It creates a set
of lists of code_list_element objects for each number of crossings to store information about each Gauss code.  These lists 
are hashed in a similar manner to the glist programme hashing for long lists to speed up the searching of codes.  Thus,
the programme creates vector of pointers to a vector of lists for each crossing number.

The first stage of gprocess is to determine whether a Gauss code contains any Reidemeister II or Reidemeister III
configurations.  If it does, it creates a list_target object that identifies the result of carrying out that Reidemeister 
move and recording the details of the move, together with any Reidemsiter I loops that are removed along the way.

The second stage of processing is to create a set of equivalence classes of code_list_element objects determined by
Reidemeister equivalence.  The class is recorded as a graph whose vertex set is the set of code_list_element objects
and whose edge set is the set of list_targets.  Within each equivalence class we create a maximal tree recorded by
noting the root of each class and for every non-root node, its parent.

The final stage of processing is to write the equivalence classes to an output file in a colon-delimited format for 
analysis by ganalyse.

February 2022

***********************************************************************************************************************/
using namespace std;
#include <fstream>
#include <iostream>
#include <sstream>
#include <cstdlib>
#include <cstring>
#include <vector>
#include <list>
#include <algorithm>
#include <csignal>
#include <iomanip>
#include <ctype.h>
#include <functional>

/******************** Global Variables **********************/
bool 			UNORIENTED_SEARCH = false;
bool 			WRITE_ELEMENT_POINTERS = false;
bool 			WRITE_EQUIVALENCE_CLASSES = false;
bool 			RETAIN_CLASS_IDENTIFIERS = false;
bool			sigusr1_request = false;
bool			termination_request = false;
int				max_num_crossings=0;
int				num_components=0;
int 			hash_threshold=100;
int 			shift_power = 23;
string			filepath;

bool     		DEVELOPMENT_MODE = false;

ofstream 	output;
extern ofstream    debug;

#include <glist.h>

/********************* Function prototypes ***********************/
void set_programme_long_option(char* cptr, string source);
void set_programme_short_option(char* cptr);
void print_prog_params (ostream& os, string level, string prefix);
int Reidemeister_I_present(gauss_orientation_data& gauss_data);
void remove_Reidemeister_I(generic_code_data& code_data, list_target& target);
bool classical_Reidemeister_II_present(gauss_orientation_data& gauss_data, pair<int,int>& bigon);
void locate_code(list_target& target, vector<list<code_list_element> >& code_list, bool hashed, string preferred_moved_code, int num_crossings);
code_list_element* find_next_root(vector<vector<list<code_list_element> >*>& code_list_pointer);
void check_debug_option_parameters(char* pptr, string option);
bool debug_setup(char* argv_ptr);
void debug_help();
void helpinfo();
string gauss_line_reflect (string c, bool unoriented);
string gauss_plane_reflect (string c, bool unoriented);
void reverse_root_path(code_list_element* tree_node, code_list_element* new_parent);
void equivalence_class_location_error(int class_identifier, string message);
generic_code_data remove_Reidemeister_II(gauss_orientation_data gauss_data, pair<int,int>& bigon, list_target& t);
string minimum_class_representative(vector<list<code_list_element> >& code_list, list<equivalence_class>& class_list, int num_crossings, bool hashed, string s);

namespace util
{
	string itos(long n);
}

void sigterm_handler (int sig) // kill -15, or just kill
{
	cout << "\ngprocess programme received SIGTERM, setting termination_request=true" << endl;
	termination_request = true;
}

/* sigusr1_handler causes gprocess to report progress on the current task and continue. */
void sigusr1_handler (int sig) // kill -10
{
	cout << "\ngprocess programme received SIGUSR1, setting sigusr1_request=true" << endl;
	sigusr1_request = true;
}


/******************* Main Function ************************/

int main (int argc, char* argv[])
{	
	signal(SIGTERM, sigterm_handler); // register the SIGTERM handler
	signal(SIGUSR1, sigusr1_handler); // register the SIGUSR1 handler
	
	/* take a copy of the original command line in case we're debugging and remove
	   the debug details.  We will want to record the original command line for the debug file.
	*/
	
	vector<string> original_programme_arguments(argc); 
	for (int i = 0; i< argc; i++)
		original_programme_arguments[i] = string(argv[i]);
	
	ifstream    input;

	ofstream 		tempfile;

	time_t start_time = time(0);
	std::tm* start_timeinfo = localtime (&start_time);
	
    /* Read the programme parameters and input files */
	for (int i=1; i < argc; i++)
	{
		if (*(argv[i]) == '-')
		{
			if (*(argv[i]+1) == '-')
				set_programme_long_option(argv[i], "command line");
			else 
				set_programme_short_option(argv[i]);
		}
	}	
	
if (debug_control::DEBUG >= debug_control::SUMMARY)
	print_prog_params (debug, "basic", "");

	if (!DEVELOPMENT_MODE && (!max_num_crossings || !num_components ))
	{
		cout << "\n";
		if (!max_num_crossings)
			cout << "max-crossings option required" << endl;
			
		if (!num_components)
			cout << "num-cpts option required" << endl;
			
		helpinfo();
	}

if (debug_control::DEBUG >= debug_control::SUMMARY)
{
	debug << "full command line: ";
	for (int i = 0; i< argc; i++)
		debug << original_programme_arguments[i] << ' ';
	debug << endl;
}


/*DEVELOPMENT_MODE is used for testing individual functions during development */
if (DEVELOPMENT_MODE)
{
	exit(0);
	/**************/
}
/* End of DEVELOPMENT_MODE code */
		
if (debug_control::DEBUG >= debug_control::SUMMARY)
{
	debug << "main: max_num_crossings = " << max_num_crossings << endl;
	debug << "main: num_components = " << num_components << endl;
}
	
    output.open ("gprocess.out");

    if (!output)
    {
		cout << "\nError opening output file\n";
		exit(0);
	}
	else
	{
		output << "Output from gprocess ";
		for (int i=1; i< argc; i++	)
			output << argv[i] << " ";
		output << "\n";
		
		output << "Start time: " << asctime(start_timeinfo);
	}

	generic_code_data code_data;
			
	input_control::ACCEPT_MAP |= input_control::gauss_code;	
	
if (debug_control::DEBUG >= debug_control::SUMMARY)
	debug << "main: input_control::ACCEPT_MAP set to 0x" << hex << input_control::ACCEPT_MAP << dec << endl;

	int code_list_length = 1 << shift_power;

if (debug_control::DEBUG >= debug_control::SUMMARY)
	debug << "code_list_length = " << code_list_length << endl;

	/* an array of pointers to code list vectors, allowing us to index using the actual number of crossings */
	vector<vector<list<code_list_element> >*> code_list_pointer(max_num_crossings+1);	

	vector<int> num_codes(max_num_crossings+1);
	int min_crossings = (num_components == 1? 2 : num_components-1);
	for (int num_crossings = min_crossings; num_crossings <= max_num_crossings; num_crossings++)
	{	
		int num_signs = 1 << num_crossings;
		vector<int> sign(num_crossings);		
		string input_file = filepath+"/glist-"+util::itos(num_crossings);
		ifstream input(input_file.c_str());

if (debug_control::DEBUG >= debug_control::SUMMARY)
	debug << "gauss: read " << input_file << " file" << endl;
		
		if(!input)
		{
			cout << "\nError opening input file " << input_file << endl;
			exit(0);
		}	
		
		string next_line;
		int index = 0;


		while(getline(input,next_line))
		{
			if ((next_line[0] == 'O' || next_line[0] == 'U') && next_line[1] == '1')
			{				
				/* read in an unsigned Gauss code and alocate crossings signs in all combinations */
				Gauss_code unsigned_code(next_line, 2*num_crossings, num_components);	

if (debug_control::DEBUG >= debug_control::DETAIL)
{
	debug << "gauss: next unsigned code ";
	write_gauss_code(debug,unsigned_code);
	debug << endl;
}				
				
				for (int i=0; i< num_signs; i++)
				{
					for (int j=0; j<num_crossings; j++)
					{
						if (i & (1 << j) ) // if the jth bit of i is set it's a positive crossing otherwise negative
							sign[num_crossings-j-1] = 1;
						else
							sign[num_crossings-j-1] = -1;						
					}

if (debug_control::DEBUG >= debug_control::EXHAUSTIVE)
{
	debug << "gauss: i = " << i << ", sign = ";
	for (int j=0; j< num_crossings; j++)
		debug << sign[j] << ' ';
	debug << endl;
}				

					Gauss_code signed_code=unsigned_code;
					for (int j=0; j< 2*num_crossings; j++)
						signed_code.code[j] += sign[signed_code.code[j]/6];

if (debug_control::DEBUG >= debug_control::DETAIL)
{
	ostringstream oss;
	write_gauss_code(oss, signed_code);
	debug << "gauss: next signed code " << oss.str() << endl;
}

					generic_code_data code_data;
					ostringstream oss;
					write_gauss_code(oss, signed_code);
					read_gauss_code(code_data,oss.str());
					
					if (debug_control::DEBUG >= debug_control::SUMMARY)
						debug << "gauss: next signed code " << oss.str() << endl;
					
					string preferred_gauss_code = over_preferred_gauss_code(code_data,UNORIENTED_SEARCH);
											
					if (num_codes[num_crossings] < hash_threshold)
					{				
						code_list_element e(num_crossings,index,0);
						e.code = preferred_gauss_code;
						
if (debug_control::DEBUG >= debug_control::SUMMARY)
	debug << "gauss: e.code " << e.code << endl;
						
						bool distinct = true;						
						for (int j=0; j < index; j++)
						{
							list<code_list_element>& hash_list = (*code_list_pointer[num_crossings])[j];
							if (hash_list.begin()->code == e.code)
							{
								distinct = false;
								break;
							}
						}
						
						if (distinct)
						{
							list<code_list_element>& hash_list = (*code_list_pointer[num_crossings])[index];
							hash_list.push_back(e);
							index++;
						}
						else
						{
if (debug_control::DEBUG >= debug_control::EXHAUSTIVE)
{
	debug << "gauss: duplicate code_list_element ";
	write_code_list_element(debug,e,-1);
	debug << endl;
}
						}
					}
					else
					{
						size_t hash_val = hash<string>()(preferred_gauss_code);
						hash_val >>= 64-shift_power; // the hash is 64 bits long
	
						list<code_list_element>& hash_list = (*code_list_pointer[num_crossings])[hash_val];
						code_list_element e(num_crossings,hash_val,hash_list.size());
						e.code = preferred_gauss_code;
						
if (debug_control::DEBUG >= debug_control::SUMMARY)
	debug << "gauss: e.code " << e.code << endl;
						
						if (find_if(hash_list.begin(), hash_list.end(), code_equals(e.code)) == hash_list.end())
							hash_list.push_back(e);
						else
						{
if (debug_control::DEBUG >= debug_control::EXHAUSTIVE)
{
	debug << "gauss: duplicate code_list_element ";
	write_code_list_element(debug,e,-1);
	debug << endl;
}
						}							
					}																		
				}
			}
			else if (next_line.find("generated") != string::npos)
			{
				istringstream iss(next_line.substr(10));
				iss >> num_codes[num_crossings];
				
				/* multiply up the number of unsigned codes by th enumber of signs */
				num_codes[num_crossings] *= (1 << num_crossings);
				
if (debug_control::DEBUG >= debug_control::SUMMARY)
	debug << "read distinct_code_count = " << num_codes[num_crossings] << " from " << input_file << ": " << next_line.substr(10) << endl;
	
				if (num_codes[num_crossings] < hash_threshold)
					code_list_pointer[num_crossings] = new vector<list<code_list_element> >(num_codes[num_crossings]);
				else
					code_list_pointer[num_crossings] = new vector<list<code_list_element> >(code_list_length);
			}
		}
	}

if (debug_control::DEBUG >= debug_control::SUMMARY)
{
	debug << "code_lists:" << endl;
	for (int i = min_crossings; i <= max_num_crossings; i++)
	{
		debug << i << ": address " << hex << code_list_pointer[i] << dec << " size "  << code_list_pointer[i]->size() << endl;
	}
}

	/* update the code_list_elements for each list to indicate where codes are related by Reidemeister moves 
	
	   The list_targets correspond to the edges in graph of equivalent codes.  If there is a Reidemeister III 
	   move between two codes then there will be a list-target from each node pointing to the other.
	   If there is a Reidemeister I or II move between two codes, it will only be detected in one direction, so we add
	   a reverse list_target with no moves in the opposite direction so that every vertex in the graph has a record of all
	   its incident edges.

	*/	

	cout << "create Reidemeister equivalence graph" << endl;
	
	int element_list_count=0; // used to index the preferred codes in code_list
	for (int num_crossings = min_crossings; num_crossings <= max_num_crossings; num_crossings++)
	{	
		
		int list_length = code_list_length;
		if (num_codes[num_crossings] < hash_threshold)
			list_length = num_codes[num_crossings];
		
		for (int i=0; i< list_length; i++)
		{
			list<code_list_element>::iterator lptr =  (*code_list_pointer[num_crossings])[i].begin();
			
			while (lptr != (*code_list_pointer[num_crossings])[i].end())
			{
if (debug_control::DEBUG >= debug_control::SUMMARY)
{
	debug << "checking code_list_element list of size " << (*code_list_pointer[num_crossings])[i].size() 
	      << " at location " << &(*lptr) << ", (*code_list_pointer[" << num_crossings << "])[" << i << ']' << endl;
}				
				generic_code_data code_data;
				read_gauss_code(code_data,lptr->code);

				gauss_orientation_data gauss_data(code_data);
							
							
				/* Look for a Reidemeister II configuration in the code.  We need seek only a single Reidemeister II configuration since, if r1 and r2 are
				   independent Reidemeister II moves that take a diagram A to B_1 and C_1 respectively, then we may subsequently perform r1 on C to obtain C_2 
				   and r2 on B to obtain B_2.  There may be subsequent Reidemeister II moves taking us to B_3,... and similarly C_3,...  If there is no 
				   Reidemeister III move between any of the B_i and C_j eventually we obtain B_i and C_j containing no more Reidemeister II moves and a 
				   sequence of R2 moves taking us from B_i to C_j, so we must have the same preferred Gauss code for both B_i and C_j
				   
				   24/10/22  I think the point here is that remove_Reidemeister_II will take out all of the RII configurations so, if there are more than one
				   that can be detected by classical_Reidemeister_II_present here, remove_Reidemeister_II will remove them all anyway.
				*/
				pair<int,int> bigon;
				if (classical_Reidemeister_II_present(gauss_data,bigon))
				{
if (debug_control::DEBUG >= debug_control::SUMMARY)
{
	debug << "detected Reidemeister II bigon in " << endl;
	print_code_list_element (debug, *lptr); 
	debug<< "bigon crossings " << bigon.first << " and " << bigon.second << endl;
}
					if (bigon.first > bigon.second)
						swap(bigon.first, bigon.second);
				
					list_target t; // default constructor  points to location=0 (the unknot) with no recorded Reidemeister moves
					if (num_crossings == 2 || (num_components == 1 && num_crossings == 3)) // an R2 will leave the unknot or unlink
					{
if (debug_control::DEBUG >= debug_control::SUMMARY)
	debug << "  num_crossings = " << num_crossings << ", so the R2 move leaves us with the unlink " << endl;
						/* no need to adjust the location, it is already zero; similarly, the destination string
						   correctly represents the unlink
						*/
						t.R2_move.push_back(bigon);
					}
					else
					{	
						/* remove_Reidemeister_II will add Reidemeister moves to t */
						generic_code_data reduced_code_data = remove_Reidemeister_II(gauss_data,bigon,t);
						
						bool look_for_target = true;
						
						if (reduced_code_data.num_crossings == 0)
						{
if (debug_control::DEBUG >= debug_control::SUMMARY)
	debug << "  The R2 move reduces the code to zero crossings " << endl;
							/* don't need to do anything since t already points to the unknot (location = 0, destination = "") */
							look_for_target = false;
						}
						else if (reduced_code_data.num_components < num_components)
						{							
							/* we have separated the diagram into reduced_code_data and the unknot */
if (debug_control::DEBUG >= debug_control::SUMMARY)
{
	debug << "  The R2 move separates the code into two components ";
	write_gauss_code(debug,reduced_code_data);
	debug << " and the unknot" << endl;
}
							look_for_target = false;
							
							/* the location is already zero, set the destination to the concatenation of the two codes. */
							ostringstream oss;
							write_gauss_code(oss,reduced_code_data);
							oss << " && unknot";
							t.destination = oss.str();							
						}
						else if (num_components > 1)
						{
							vector<int> dummy_component_flags;
							generic_code_data subsequent_code_data = partition_peer_code(reduced_code_data,dummy_component_flags);
							if (subsequent_code_data.num_crossings != 0)
							{
if (debug_control::DEBUG >= debug_control::SUMMARY)
{
	debug << "  The R2 move separates the code into two components ";
	write_gauss_code(debug,reduced_code_data);
	debug << " and "; 
	write_gauss_code(debug,subsequent_code_data);
	debug << endl;
}
								look_for_target = false;
								/* the location is already zero, set the destination to the concatenation of the two codes. */
								ostringstream oss;
								write_gauss_code(oss,reduced_code_data);
								oss << " && ";
								write_gauss_code(oss,subsequent_code_data);
								t.destination = oss.str();							
							}
							else
							{
if (debug_control::DEBUG >= debug_control::SUMMARY)
	debug << "  The R2 move does not separate the code, look for the target code in the lists";
							}
						}
						
						if (look_for_target)
						{
							gauss_orientation_data R2_reduced_gauss_data(reduced_code_data);
							string preferred_reduced_code = over_preferred_gauss_code(reduced_code_data,UNORIENTED_SEARCH);							
													
							locate_code(t, *code_list_pointer[reduced_code_data.num_crossings], (num_codes[reduced_code_data.num_crossings] >= hash_threshold), preferred_reduced_code,reduced_code_data.num_crossings);
						}
					}

					lptr->num_targets = 1;
					lptr->target = {t};

if (debug_control::DEBUG >= debug_control::SUMMARY)
	debug << "  R2 move target is location " << t.location << endl;
					
					if (t.location != 0)
					{
						
						/* push a reverse list_pointer onto the list_target list at t.location containing no moves */
						list_target reverse_target(&(*lptr));
						t.location->target.push_back(reverse_target);
						t.location->num_targets++;
						
if (debug_control::DEBUG >= debug_control::SUMMARY)
	debug << "  added a reverse list target from " << t.location << " to " << reverse_target << endl;
	
					}					
				}
				else
				{
if (debug_control::DEBUG >= debug_control::SUMMARY)
	debug << "no R2 configurations found" << endl;
				}
				
				Reidemeister_III_return R3_return;
				
				if (num_crossings >2)
				{
if (debug_control::DEBUG >= debug_control::SUMMARY)
{
	debug << "looking for R3 moves in " << lptr->code << endl;
}						
					R3_return = Reidemeister_III_present (code_data);
				}

				if (R3_return.Reidemeister_III_found)
				{
					int num_moves = R3_return.R3_configuration.numrows();
					
if (debug_control::DEBUG >= debug_control::SUMMARY)
	debug << "found " << num_moves << " R3 configurations " << endl;
	
					vector<list_target> move_target(num_moves);					
					matrix<int>& R3_moves = R3_return.R3_configuration;

if (debug_control::DEBUG >= debug_control::SUMMARY)
{
	debug << "R3 configuration: " << endl;
	print(R3_moves,debug,3,"");
	debug << endl;
}
					
					/* create the list of targets for each Reidemeister III configuration, we use num_distinct_moves as an index into move_target 
					   and as a counter
					*/
					int num_distinct_moves = 0;
					bool distinct;
					for (int j=0; j< num_moves; j++)
					{
						
						generic_code_data moved_code = Reidemeister_III_move(code_data, R3_moves[j][0], R3_moves[j][1], R3_moves[j][2]);
						vector<int> Reidemeister_move = {R3_moves[j][0], R3_moves[j][1], R3_moves[j][2], R3_moves[j][3], R3_moves[j][4], R3_moves[j][5]};

if (debug_control::DEBUG >= debug_control::SUMMARY)
{
	debug << "  Reidemeister move: ";
	for (size_t p=0; p<Reidemeister_move.size(); p++)
		debug << Reidemeister_move[p] << ' ';
	debug << endl;
}

						list_target next_target;
						next_target.R3_move.push_back(Reidemeister_move);
						
if (debug_control::DEBUG >= debug_control::SUMMARY)
{
	debug << "  moved code: ";
	write_gauss_code(debug,moved_code,false);
	debug << endl;
}						
						remove_Reidemeister_I(moved_code, next_target);	

						distinct = true;
						if (moved_code.num_crossings == 0)
						{							
							/* have we already found a target to the unknot? */
							if (lptr->num_targets == 1)
							{
								/* there was a Reidemeister II configuration */
								if (lptr->target[0].location == 0)
									distinct = false;
							}
							
							for (int k=0; distinct && k< num_distinct_moves; k++)
							{
								if (move_target[k].location == 0)
								{
									distinct = false;
									break;
								}
							}
						}
						else
						{
							string preferred_moved_code = over_preferred_gauss_code(moved_code,UNORIENTED_SEARCH);							
							
if (debug_control::DEBUG >= debug_control::SUMMARY)
	debug << "preferred_moved_code " << preferred_moved_code << endl;
	
							locate_code(next_target,*code_list_pointer[moved_code.num_crossings], (num_codes[moved_code.num_crossings] >= hash_threshold), preferred_moved_code, moved_code.num_crossings);
							
							if (lptr->num_targets == 1)
							{
								/* there was a Reidemeister II configuration */
								if (lptr->target[0].location == next_target.location)
									distinct = false;
							}
							
							for (int k=0; distinct && k< num_distinct_moves; k++)
							{
								if (move_target[k].location == next_target.location)
								{
									distinct = false;
									break;
								}
							}
							
						}
						
						if (distinct)
						{							
if (debug_control::DEBUG >= debug_control::SUMMARY)
{
	debug << "identified as a distinct move to location ";
	if (next_target.location == 0)
		debug << "0, the unlink" << endl;
	else
		debug << next_target.location << " having " << next_target.location->num_crossings << " crossings" << endl;
}

							move_target[num_distinct_moves] = next_target;							
							num_distinct_moves++;
							
							if (next_target.location != 0 && next_target.location->num_crossings < num_crossings)
							{
								/* push a reverse list_pointer onto the list_target list at next_target.location containing no moves */
								list_target reverse_target(&(*lptr));
								next_target.location->target.push_back(reverse_target);
								next_target.location->num_targets++;
							}							
						}
						else
						{
if (debug_control::DEBUG >= debug_control::SUMMARY)
	debug << "not a distinct move, already visited location " << (next_target.location?next_target.location: 0) << endl;
						}
					} // end of for (int j=0; j< num_moves; j++)


					lptr->num_targets += num_distinct_moves;
					
					for (int i=0; i< num_distinct_moves; i++)
						lptr->target.push_back(move_target[i]);
					
				}
				else
				{
if (debug_control::DEBUG >= debug_control::SUMMARY)
	debug << "no R3 configurations found" << endl;
					
					/* just record the index for this code in its corresponding list_element */
				}

				lptr->index = element_list_count++;
			
				if (sigusr1_request)
				{
					/* report progress */
					cout << "\nworking on list for " << num_crossings << " crossings, progress " << element_list_count << " of " << num_codes[num_crossings] << endl;
					sigusr1_request = false;
				}
								
				lptr++;
			} // end of while (lptr != (*code_list_pointer[num_crossings])[i].end())
		}
		
		if (WRITE_ELEMENT_POINTERS)
		{
			/* write the Reidemeister III data to the output file */
			output << "\n" << num_codes[num_crossings] << " code_list_elements for " << num_crossings << " crossings" << endl;
			for (int i=0; i< list_length; i++)
			{
				list<code_list_element>::iterator lptr =  (*code_list_pointer[num_crossings])[i].begin();
				
				if ((*code_list_pointer[num_crossings])[i].size() != 0)
				{
				
if (debug_control::DEBUG >= debug_control::SUMMARY)
	debug << "code_list entry at index " << i << ", element list  size = " << (*code_list_pointer[num_crossings])[i].size() << endl;
	
					while (lptr != (*code_list_pointer[num_crossings])[i].end())
					{
if (debug_control::DEBUG >= debug_control::SUMMARY)
{
	debug << "code_list_element " << endl;
	print_code_list_element (debug, *lptr); 
}
	
	
						output << lptr->index << ". ";
						
						if (list_length == code_list_length)
							output << "place = " << i << ": ";
						
						print_code_list_element(output,*lptr);
							
						lptr++;
					}
				}
			}
		}
	}

if (debug_control::DEBUG >= debug_control::SUMMARY)
	debug << "code_list contains a total of " << element_list_count << ", code_list_element objects" << endl;
	
	/* We now create a set of set of equivalence classes of pointers to the code_list_elements in the lists referenced by the 
	   code list pointers, based on the corresponding Gauss codes being equivalent up to Reidemeister moves.  The list_targets
	   tell us how pairs of code_list_elements are related by single moves.
	   
	   Equivalence clases are represented as connected graphs, the vertices of which are pointers to code_list_elements and the edges are 
	   list_targets.  Within each graph we maintain a maximal tree rooted at the code-list_element having the minimal representation
	   Gauss code for that class.  We record with each vertex in the graph it's parent vertex in the maximal tree.  Each vertex in the 
	   equivalence class graph has an index (code_list_element.index) representing the index of the vertex amongst the set of vectors
	   vector<vector<list<code_list_element> >*> code_list_pointer; i.e. a global index across all of the code_list_pointers that have
	   been created.
	   
	   The equivalence classes are identified by a set of one or more class_identifiers and code_list_elements have their
	   member_equivalence_class set to the class_identifier to which they have been determined to belong.  
	   
	   The equivalence classes are constructed from a sequence of trees corresponding to each class identifier.  Working from codes with the 
	   smallest number of crossings first, we choose the 
	   first code_list_element whose member_equivalence_class has not been set as the starting point for a breadth first search 
	   to determine a maximal "class-tree" for the corresponding class identifier.  The starting point is referred to as the "local root", 
	   since ultimately the root of the tree will be the code_list_element in the equivalence class having the minimal gauss code, 
	   where we use the same comparison of gauss codes as used in the determination of the over_preferred_gauss_code (so it is 
	   dependent on whether this is an unoriented search or not.) We track the best (i.e. minimal) code we have encountered as the class-tree
	   is created.
	   
	   As we extend the tree, we note for each vertex, the parent vertex in the unique path from that vertex to the local root.
	   
	   If we extend the class-tree to its maximum extent without encountering a code_list_element previously assigned to a class, we
	   re-orient the tree so that the root is the "minimal" code_list_element described above.  Re-orienting a tree requires only that 
	   the parent-child relationship be reversed on the unique path from the new root to the old root.
	   
	   If we encounter a code_list_element previously assigned to a class as we extend a class-tree, we compare the current minimal code for the 
	   class-tree being constructed, with the root code of the class we have encountered.  The two classes will be amalgamated and the minimal
	   code of these two values will be used the root.  Since we may encounter multiple different equivalence classes as we extend the class-tree,
	   and we may encounter a previous class more than once, we form a list of graft_data objects that identify how an earlier class-tree and the current
	   class-tree should be grafted.
	   
	   CURRENTLY WE SPLICE TREES AT THE FIRST ENCOUNTER, RATHER THAN SEEKING AN OPTIMAL SPLICE POINT, SINCE IT IS UNCLEAR HOW TO COMPARE POSSIBLE GRAFT LOCATIONS.
	   
	   There will only be one graft_data object with the current class as the from_class, since at the end of the process of extending the class-tree there will be
	   a unique root.  There may be multiple graft_data objects with the current class as the to_class.  Note that if we splice trees together there is a single
	   equivalence_class object for the amalgamated trees.	   
	*/
	list<equivalence_class> class_list;
	
	code_list_element* local_root = find_next_root(code_list_pointer);
	code_list_element* minimal_root;
	bool finished = false;
	int class_count = 1;
	int total_node_count = 0;  // for progress reporting

	cout << "create equivalence classes" << endl;

	do
	{
		if (sigusr1_request)
		{
			/* report progress */
			cout << "\nidentified " << class_count << " equivalence classes involving a total of " << total_node_count << " nodes" << endl;
			sigusr1_request = false;
		}

if (debug_control::DEBUG >= debug_control::SUMMARY)
	debug << "equivalence class " << class_count << endl;

cout << "\r" << class_count << flush;

		local_root->member_equivalence_class = class_count;
		local_root->parent = 0;
				
		equivalence_class next_class;
		next_class.root = local_root;
		next_class.min_representation = local_root->code;
		next_class.min_num_crossings = local_root->num_crossings;
		next_class.class_identifiers.push_back(class_count);

		minimal_root = local_root;


if (debug_control::DEBUG >= debug_control::SUMMARY)
	debug << "  next_root = " << next_class.root << " index " << local_root->index << " code " << next_class.min_representation << " num_targets = " << local_root->num_targets << endl;
		
		list<equivalence_class*> adjacent_classes;
		list<code_list_element*> class_target;

		graft_data local_graft; // used if we need to graft this class-tree to another tree.
		equivalence_class* local_graft_equivalence_class = 0;
		
		list<graft_data> graft_list;  // records the grafts data for grafting other trees onto this class-tree
		
		int node_count = 0;	
		class_target.push_back(local_root);		
				
		while (class_target.size() > 0)
		{
			code_list_element* next_node = *(class_target.begin());
			node_count++;
			total_node_count++;

			class_target.pop_front();

if (debug_control::DEBUG >= debug_control::SUMMARY)
	debug << "  next_node = " << next_node << ", index = " << next_node->index << ", num_targets = " << next_node->num_targets << endl;


			for (int i=0; i< next_node->num_targets; i++)
			{
				if (next_node->target[i].location == 0)
				{					
					next_class.min_representation = next_node->target[i].destination;  // this might be the empty string or a pair of disjoint codes.
					next_class.min_num_crossings = 0;
					minimal_root = next_node;
						
if (debug_control::DEBUG >= debug_control::SUMMARY)
	debug << "    next_node->target[" << i << "].location = 0 set next_class.min_representation to the empty string, minimal_root = " << minimal_root << endl;			
				}
				else
				{
if (debug_control::DEBUG >= debug_control::SUMMARY)
{
	debug << "    next_node->target[" << i << "].location " << next_node->target[i].location << ", index " << next_node->target[i].location->index 
	      << ", member_equivalence_class = " << next_node->target[i].location->member_equivalence_class << endl;
}
	
					if (next_node->target[i].location->member_equivalence_class == 0)
					{
if (debug_control::DEBUG >= debug_control::SUMMARY)
	debug << "      code " << next_node->target[i].location->code << endl;			

						/* extend the current class-tree */
						next_node->target[i].location->parent = next_node;
						next_node->target[i].location->member_equivalence_class = class_count;
						class_target.push_back(next_node->target[i].location);			

if (debug_control::DEBUG >= debug_control::SUMMARY)
{
	debug << "      set parent to " << next_node->target[i].location->parent << " equivalence class to " << next_node->target[i].location->member_equivalence_class 
	      << ", class_target.size() now "  << class_target.size() << endl;			
}
						if (next_class.change_to_representation(next_node->target[i].location->code))
						{
							next_class.min_representation = next_node->target[i].location->code;
							next_class.min_num_crossings = next_node->target[i].location->num_crossings;
							minimal_root = next_node->target[i].location;
if (debug_control::DEBUG >= debug_control::SUMMARY)
	debug << "      updated min_representation = " << next_class.min_representation << ", minimal_root updated to " << minimal_root << endl;			
						}
					}
					else if (next_node->target[i].location->member_equivalence_class != class_count)
					{
if (debug_control::DEBUG >= debug_control::SUMMARY)
	debug << "      member_equivalent_class distinct from current class identifier " << class_count << endl;			

						list<equivalence_class>::iterator ec_ptr = find_if(class_list.begin(),class_list.end(),class_identifier_includes(next_node->target[i].location->member_equivalence_class));
						if (ec_ptr == class_list.end())
							equivalence_class_location_error(next_node->target[i].location->member_equivalence_class, "looking for the class of a next_node list_target");
					
						if (&(*ec_ptr) == local_graft_equivalence_class)
						{
if (debug_control::DEBUG >= debug_control::SUMMARY)
	debug << "      current class has previously been identified as grafting to the adjacent class" << endl;			
						}
						else if (next_class.change_to_representation(ec_ptr->min_representation))
						{							
							next_class.min_representation = ec_ptr->min_representation;
							next_class.min_num_crossings = ec_ptr->min_num_crossings;
							minimal_root = next_node;

							local_graft.from_node = next_node;
							local_graft.to_node = next_node->target[i].location;								
							local_graft_equivalence_class = &(*ec_ptr);
								
if (debug_control::DEBUG >= debug_control::SUMMARY)
{
	debug << "      adjacent class has preferred representation " << endl;			
	debug << "      updated min_representation = " << next_class.min_representation << ", set local_graft from_node = " << local_graft.from_node << ", to_node = " << local_graft.to_node << endl;			
	debug << "      local_graft_equivalence_class = " << local_graft_equivalence_class << " minimal_root updated to " << minimal_root << endl;			
}
						}
						else
						{
							if (find(adjacent_classes.begin(), adjacent_classes.end(), &(*ec_ptr)) == adjacent_classes.end())
							{							
							
								/* create graft_data for the adjacent class tree */
								graft_data graft(next_node->target[i].location, next_node);
								graft_list.push_back(graft);

								/* record the fact that we've encountered this class */
								adjacent_classes.push_back(&(*ec_ptr));
							
if (debug_control::DEBUG >= debug_control::SUMMARY)
{
	debug << "    local class has preferred representation " << endl;			
	debug << "    added graft_data from " << next_node->target[i].location << ", to " << next_node << endl;			
}
							}
							else
							{
if (debug_control::DEBUG >= debug_control::SUMMARY)
	debug << "    local class has preferred representation but we have already created a graft_data object for the adjacent class" << endl;			
							}
						}
					}
					else
					{
if (debug_control::DEBUG >= debug_control::SUMMARY)
	debug << "      member_equivalent_class indistinct from current class identifier " << class_count << endl;			
					}
				}
			}
						
if (debug_control::DEBUG >= debug_control::SUMMARY)
	debug << "    after processing next_node, class_target.size() = " << class_target.size() << endl;			
		}
if (debug_control::DEBUG >= debug_control::SUMMARY)
	debug << "class contains " << node_count << " nodes, min_representation = " << next_class.min_representation << endl;			
	
			
		if (graft_list.size() != 0)
		{
			/* graft the adjacent trees as determined by the graft_list to the current tree.  The parent of the from_node becomes the to_node
			   and we reverse the parent child relationship on the path from the from_node to the root of the adjacent tree.  We append the 
			   class identifier list of the adjacent equivalent class onto the current class's list of class identifiers and remove the
			   adjacent equivalence class.
			*/
			list<graft_data>::iterator gd_ptr =  graft_list.begin();  
			
			while (gd_ptr !=  graft_list.end())
			{
if (debug_control::DEBUG >= debug_control::SUMMARY)
{
	debug << "splicing tree_node " << gd_ptr->from_node << ", index = " << gd_ptr->from_node->index << " to graft_location " << gd_ptr->to_node << ", index = " << gd_ptr->to_node->index << endl;

	debug << "initial to graft_location class_dentifiers: ";
	list<int>::iterator lptr = next_class.class_identifiers.begin();
	while (lptr != next_class.class_identifiers.end())
	{
		debug << *lptr << ' ';
		lptr++;
	}
	debug << endl;
}			

				reverse_root_path(gd_ptr->from_node,gd_ptr->to_node);

				list<equivalence_class>::iterator ec_ptr = find_if(class_list.begin(),class_list.end(),class_identifier_includes(gd_ptr->from_node->member_equivalence_class));								
				if (ec_ptr == class_list.end())
					equivalence_class_location_error(gd_ptr->from_node->member_equivalence_class, "grafting adjacent class-tree to the current tree");

				next_class.class_identifiers.splice(next_class.class_identifiers.end(),ec_ptr->class_identifiers);
		
if (debug_control::DEBUG >= debug_control::SUMMARY)
{

	debug << "spliced class_dentifiers: ";
	list<int>::iterator lptr = next_class.class_identifiers.begin();
	while (lptr != next_class.class_identifiers.end())
	{
		debug << *lptr << ' ';
		lptr++;
	}
	debug << endl;
}			
				class_list.erase(ec_ptr);					
				gd_ptr++;
			}
		}

		if (local_graft_equivalence_class != 0)
		{
			/* graft the local tree at local_graft.from_node onto the adjacent tree node identified by local_graft.to_node.  
			   The to_node becomes the parent of the from node and we reverse the parent child relationship on the path from the
			   from_node to the local_root.  We append the current class identifier to the list of class identifiers associated with the
			   adjacent tree and we do not add the equivalence class we have been creating to the list of classes.
			*/

if (debug_control::DEBUG >= debug_control::SUMMARY)
	debug << "splicing tree_node " << local_graft.from_node << ", index = " << local_graft.from_node->index << " to graft_location " << local_graft.to_node << ", index = " << local_graft.to_node->index << endl;

			reverse_root_path(local_graft.from_node,local_graft.to_node);

			/* locate the equivalence class of the to_node in the class_list */	
			list<equivalence_class>::iterator ec_ptr = find_if(class_list.begin(),class_list.end(),class_identifier_includes(local_graft.to_node->member_equivalence_class));
			if (ec_ptr == class_list.end())
				equivalence_class_location_error(local_graft.to_node->member_equivalence_class, "looking for the class of local_graft.to_node");

			ec_ptr->class_identifiers.splice(ec_ptr->class_identifiers.end(),next_class.class_identifiers);

if (debug_control::DEBUG >= debug_control::SUMMARY)
{
	debug << "spliced class_dentifiers: ";
	list<int>::iterator lptr = ec_ptr->class_identifiers.begin();
	while (lptr != ec_ptr->class_identifiers.end())
	{
		debug << *lptr << ' ';
		lptr++;
	}
	debug << endl;
}			
		}
		else 
		{
if (debug_control::DEBUG >= debug_control::SUMMARY)
	debug << "  minimal_root =  " << minimal_root << ", local_root = " << local_root << endl;
	
			/* reverse the parent child relationship on the path from the minimal_root to the local_root */
			if (minimal_root != local_root)
				reverse_root_path(minimal_root,0); // new parent of minimal root is 0

			next_class.root = minimal_root;
			class_list.push_back(next_class);					
		}

		local_root = find_next_root(code_list_pointer);
		if (local_root == 0)
			finished = true;
		else
			class_count++;
			
	} while (!finished);
	
//	cout << "\ntotal of " << class_list.size() << " equivalence classes" << endl;

	/* evaluate symmetry data for the equivalence classes */
	
	cout << "\ncreate reflection symmetry data for non-trivial classes" << endl;
	
	/* remove line, plane and combined reflections from the non-trivial classes */
if (debug_control::DEBUG >= debug_control::SUMMARY)
	debug << "removing plane reflections from non-trivial classes" << endl;
	
	list<equivalence_class>::iterator lptr1 = class_list.begin();
	int class_index = 0;
	do
	{
		if (lptr1->min_representation == "" || lptr1->min_representation.find(" && ") != string::npos)
		{
			lptr1++;
			continue;
		}
			
cout << "\r" << ++class_index << flush;
				
		bool found = false;
			
if (debug_control::DEBUG >= debug_control::SUMMARY)	
	debug << "check symmetries for " << lptr1->min_representation;

		/* plane reflection interchanges both over and under-crossings and crossing signs, the order in which crossings are encountered remains unchanged
		   line reflection interchanges crossing signs but over and under-crossings remain unchanged, the order in which crossings are encountered remains unchanged.
		   the composition of plane and line reflection interchanges over and under-crossings but crossing signs remain unchanged.
		*/
		int num_crossings = lptr1->min_num_crossings;
		bool hashed = (num_codes[num_crossings] >= hash_threshold);
		string line_reflected_code = minimum_class_representative(*code_list_pointer[num_crossings],class_list, num_crossings, hashed, gauss_line_reflect(lptr1->min_representation,UNORIENTED_SEARCH));
		string plane_reflected_code = minimum_class_representative(*code_list_pointer[num_crossings],class_list, num_crossings, hashed, gauss_plane_reflect(lptr1->min_representation,UNORIENTED_SEARCH));
		string double_reflected_code = minimum_class_representative(*code_list_pointer[num_crossings],class_list, num_crossings, hashed, gauss_line_reflect(plane_reflected_code,UNORIENTED_SEARCH));

		list<equivalence_class>::iterator lptr2 = lptr1;
		lptr2++;
		list<equivalence_class>::iterator mark;
					
		while (lptr2 != class_list.end())
		{					

		if (lptr2->min_representation == "")
		{
			lptr2++;
			continue;
		}
			
		if (lptr2->min_representation == line_reflected_code) 
		{
			lptr1->symmetry_type = equivalence_class::symmetry::line_reflect;
			mark = lptr2;
			found = true;
			break;
				
if (debug_control::DEBUG >= debug_control::SUMMARY)
	debug << ": found line reflected class " << distance(class_list.begin(),lptr2) << ", min_representation " << lptr2->min_representation << endl;					
		}			
		else if (lptr2->min_representation ==  plane_reflected_code)
		{
			lptr1->symmetry_type = equivalence_class::symmetry::plane_reflect;
			mark = lptr2;
			found = true;
			break;
								
if (debug_control::DEBUG >= debug_control::SUMMARY)
	debug << ": found plane reflected class " << distance(class_list.begin(),lptr2) << ", min_representation " << lptr2->min_representation << endl;					
		}			
		else if (lptr2->min_representation == double_reflected_code)
		{
			lptr1->symmetry_type = equivalence_class::symmetry::double_reflect;
			mark = lptr2;
			found = true;
			break;
				
if (debug_control::DEBUG >= debug_control::SUMMARY)
	debug << ": found double reflected class " << distance(class_list.begin(),lptr2) << ", min_representation " << lptr2->min_representation << endl;					
		}			
			
			lptr2++;
			
		} //while (lptr2 != class_list.end())

		if (found)
		{
			lptr1->symmetry_representative = distance(class_list.begin(),mark);
			list<int>& class_symmetry_list = mark->symmetry_class;
			class_symmetry_list.push_back(distance(class_list.begin(),lptr1));
			list<int>::iterator lptr = lptr1->symmetry_class.begin();
			while(lptr != lptr1->symmetry_class.end())
			{
				class_symmetry_list.push_back(*lptr);
				lptr++;
			}
		}
		else
		{
if (debug_control::DEBUG >= debug_control::SUMMARY)
	debug << ": not found" << endl;				
		}

		lptr1++;
		
	} while (lptr1 != class_list.end());		

	/* end of symmetry data evaluation */
	
	
	list<equivalence_class>::iterator cptr = class_list.begin();
	
	vector<bool> visited(element_list_count,false);
	int non_trivial_class_count = 0;
	
	output << "\nCLASSCOUNT: " << class_list.size() << endl;
	
	/* write the equivalence classes to the output file as a colon delimited format.  We use index -1 to indicate the unknot and the
	   root of an equivalence class.

		default:
			class: <-1>,<min-gauss-code-representation>,<class-identifier>,<symmetry-data>
		RETAIN_CLASS_IDENTIFIERS:	
			class: <-1>,<min-gauss-code-representation>,<num-class-identifiers>,{<class-identifiers>},<symmetry-data>
		node:  <index>,<parent-index>,<gauss-code>,<class-identifier>,<num-list-targets>,{<list-target>}

		where: <parent-index> is set to -1 if the node is the root of a class-tree
		       {<class-identifiers>} is a comma delimited set of integers
		       {<list-target>} is a comma delimted set of list-targets.
		       
		Each <list-target> has the form:
		
		<index>,<num-R1-moves>,<num-R2-moves>,<num-R3-moves>,{<R1-move>},{<R2-move>},{<R3-move>}
		
		where: <index> is set to -1 if the list target location is zero
		       <R1-move> = list of integers
		       <R2-move> = list of integer pairs
		       <R3-move> = list of integer triples (<ingress over-arc-edge> <ingress under-arc-edge-1> <ingress under-arc-edge-2>)	
		
		By default, we re-assign the class-identifiers so that each class has a single class identifier and each code_list_element 
		belonging to that class has the same member_equivalence_class.  If the boolean RETAIN_CLASS_IDENTIFIERS is true we retain the
		original class identifiers for debugging purposes.
	*/
	class_index = (RETAIN_CLASS_IDENTIFIERS?-1:0);
	
	output << "NODECOUNT: " << element_list_count << endl;
	while (cptr != class_list.end())	
	{
cout << "\r" << class_index << flush;

if (debug_control::DEBUG >= debug_control::SUMMARY)
	debug << "write equivalence class " << class_index << ", root index = " << cptr->root->index << endl;

		visited[cptr->root->index] = true;

		
		if (WRITE_EQUIVALENCE_CLASSES)
			output << -1 << ':';
			
		if (cptr->min_representation == "")
		{
			if (WRITE_EQUIVALENCE_CLASSES)
				output << "E";
		}
		else
		{
			if (WRITE_EQUIVALENCE_CLASSES)
				output << cptr->min_representation;
				
			if (cptr->min_representation.find(" && ") == string::npos)
				non_trivial_class_count++;
		}

		if (WRITE_EQUIVALENCE_CLASSES)
		{
			if (RETAIN_CLASS_IDENTIFIERS)
			{
				list<int>::iterator lptr = cptr->class_identifiers.begin();
		
				output << ':' << cptr->class_identifiers.size() << ':';
					
				while (lptr != cptr->class_identifiers.end())
				{
					output << *lptr;			
					lptr++;
					
					if (lptr != cptr->class_identifiers.end())
						output << ':';
				}
			}
			else
			{
				output << ':' << class_index;

if (debug_control::DEBUG >= debug_control::SUMMARY)
{
	debug << "replacing class identifiers ";
	list<int>::iterator lptr = cptr->class_identifiers.begin();
	while (lptr != cptr->class_identifiers.end())
	{
		debug << *lptr << ' ';			
		lptr++;
	}
	debug << " with class_index " << class_index << endl;
}

			}

			/* Write the non-trivial equivalence class symmetry data to output in the colon delimeted form:
			
			   if symmetry-type == none	   
			     <class-identifier>:<symmetry-type>:<symmetry-class-length>:<symmetry-class>
			   if symmetry-type != none	   
			     <class-identifier>:<symmetry-type>:<symmetry_representative>:<symmetry-class-length>:<symmetry-class>
			     
			   where <symmetry-class> = <element_first>:...:<element_last>
			*/

			if (cptr->min_representation != "")
			{	
				output << ':' << cptr->symmetry_type << ':';
				
				if (cptr->symmetry_type != equivalence_class::symmetry::none)
					output << cptr->symmetry_representative << ':';
					
				output << cptr->symmetry_class.size();
				
				if (cptr->symmetry_class.size() != 0)
				{
					list<int>::iterator lptr = cptr->symmetry_class.begin();
					while (lptr != cptr->symmetry_class.end())
					{
						output << ':' << *lptr;
						lptr++;
					}
				}
			}
			
			output << endl;
			write_code_list_element(output,*(cptr->root),class_index);
		}

		if (cptr->root->num_targets != 0)
		{			
			
			list<code_list_element*> class_target;
			
			for (int i=0; i< cptr->root->num_targets; i++)
			{
				if (cptr->root->target[i].location != 0)
				{
					
if (debug_control::DEBUG >= debug_control::SUMMARY)
	debug << "    add node " << cptr->root->target[i].location->index << " to class_target " << endl;
	
					class_target.push_back(cptr->root->target[i].location);					
					visited[cptr->root->target[i].location->index] = true;
				}
			}
				
			while (class_target.size() > 0)
			{
				code_list_element* next_node = *(class_target.begin());
				
if (debug_control::DEBUG >= debug_control::SUMMARY)
	debug << "  next node " << next_node->index << endl;
				
				if (WRITE_EQUIVALENCE_CLASSES)
					write_code_list_element(output,*next_node,class_index);

				class_target.pop_front();
				
				for (int i=0; i< next_node->num_targets; i++)
				{
					/* add the target location to class_target if it is not the unlink, has not already been written and is not already on the list pending being written */
					if (next_node->target[i].location != 0 && !visited[next_node->target[i].location->index])
					{

if (debug_control::DEBUG >= debug_control::SUMMARY)
	debug << "    add node " << next_node->target[i].location->index << " to class_target " << endl;

						class_target.push_back(next_node->target[i].location);	

						visited[next_node->target[i].location->index] = true;
					}
				}				
			}				
		}	
		
		cptr++;
		if (!RETAIN_CLASS_IDENTIFIERS)
			class_index++;
	}

	cout << "\n" << non_trivial_class_count << " non-trivial classes" << endl;
	
if (debug_control::DEBUG >= debug_control::SUMMARY)
	debug << "non_trivial_class_count = " << non_trivial_class_count << endl;
	
	time_t end_time = time(0);
	double time_exp = difftime(end_time, start_time);
	int dd = static_cast<int>(time_exp/3600/24);
	time_exp -= dd*3600*24;
	int hh = static_cast<int>(time_exp/3600);
	time_exp -= hh*3600;
	int mm = static_cast<int>(time_exp/60);
	time_exp -= mm*60;
	int ss = static_cast<int>(time_exp);
	output << "\nRunning time: " << dd << " days " << hh << " hours " << mm	<< " minutes " << ss << " seconds" << endl;


	for (int i=min_crossings; i<max_num_crossings+1; i++)
	{

if (debug_control::DEBUG >= debug_control::SUMMARY)
	debug << " deleting code_list_pointer " << hex << code_list_pointer[i] << endl;

		delete code_list_pointer[i];
	}

	output.close();
	cout << endl;

	return 0;
} /* End of main */

/***************  Functions required for setting up gprocess specific debug **************/

void set_main_debug_option_parameter(char* pptr, string option);
void set_main_debug_option(char* start, char* end)
{
	char  loc_buf[end-start+2];
	char* c1 = start;
	char* c2 = loc_buf;

	/* if both start and end are zero, display debug help information.
	   this has been included here in this manner so that each time a debug option
	   is added, the help will be updated (hopefully!)
	*/
	if (start == 0 && end == 0)
	{
		set_main_debug_option_parameter(0,"gprocess");
		return;
	}

	do
	{
		*c2++ = *c1++;
	} while (c1 <= end);
	
	*c2 = '\0';

	char* pptr = strchr(loc_buf,'{');
	if (pptr)
	{
		*pptr++ = '\0';
	}

	/* now, even if there are parameters, the first part of loc_buf 
	   is a C-string that identifies the option */
	
	if (!strcmp(loc_buf,"gprocess"))
	{
		debug_control::DEBUG = debug_control::SUMMARY;
		debug << "set_main_debug_option: setting debug option debug_control::DEBUG = debug_control::SUMMARY" << endl;		
		
		if (pptr)
		{
			/* check for any parameters */
			check_debug_option_parameters(pptr, "gprocess");
		}
	}

	debug.flush();

}

void helpinfo()
{
	cout << "\nUsage: gprocess [-<short-option>] [--<long-option>] ";
	cout << "\n<short-option>:";
	cout << "\n       #: debug, see #H! for details";
	cout << "\n       e: write equivalence class data to output file";
	cout << "\n       h: show normal help";
	cout << "\n       u: unoriented code lists";
	cout << "\n       H!: show this help";
	cout << "\n       #H!: display debug help screen";
	cout << "\n       p: write element pointer list to output file";
	cout << "\n<long-option>:";
	cout << "\n       filepath=<relative filepath>: specifies the location of the code lists";
	cout << "\n       hash_threshold=n: only hash codes if the code list is longer than hash threshold, default=100";
	cout << "\n       num-cpts=n: number of components";
	cout << "\n       max-crossings=n: maximum number of crossings to consider";
	cout << "\n       shift-power=n: the code_list_length is 2^n, where n is the shift-power, default=23";
	cout << endl;
	exit(0);
}

void set_main_debug_option_parameter(char* pptr, string option)
{
	if (option == "gprocess")
	{
		if (!pptr)
		{
			cout << "\t\tgprocess{summary|1:basic|2:intermediate|3:detail|4}, integer: default 0=off, no parameters sets summary" << endl;
		}
		else
		{
			if (!strcmp(pptr,"summary") || !strcmp(pptr,"1") )
			{
				debug_control::DEBUG = debug_control::SUMMARY;
				debug << "set_main_debug_option_parameter: setting debug option debug_control::DEBUG = debug_control::SUMMARY" << endl;		
			}
			if (!strcmp(pptr,"basic") || !strcmp(pptr,"2") )
			{
				debug_control::DEBUG = debug_control::BASIC;
				debug << "set_main_debug_option_parameter: setting debug option debug_control::DEBUG = debug_control::BASIC" << endl;		
			}
			else if (!strcmp(pptr,"intermediate") || !strcmp(pptr,"3"))
			{
				debug_control::DEBUG = debug_control::INTERMEDIATE;
				debug << "set_main_debug_option_parameter: setting debug option debug_control::DEBUG = debug_control::INTERMEDIATE" << endl;		
			}
			else if (!strcmp(pptr,"detail") || !strcmp(pptr,"4"))
			{
				debug_control::DEBUG = debug_control::DETAIL;
				debug << "set_main_debug_option_parameter: setting debug option debug_control::DEBUG = debug_control::SUMMARY" << endl;		
			}
			else if (!strcmp(pptr,"exhaustive") || !strcmp(pptr,"5"))
			{
				debug_control::DEBUG = debug_control::EXHAUSTIVE;
				debug << "set_main_debug_option_parameter: setting debug option debug_control::DEBUG = debug_control::EXHAUSTIVE" << endl;		
			}
		}
	}
}

void main_display_default_options()
{
	cout << "\t\tgprocess{summary}" << endl;
}

void set_programme_short_option(char* cptr)
{
	if (strchr(cptr, 'H') && strchr(cptr, '!'))
	{
		if (strchr(cptr,'#'))
		{
			debug_help();
		}
		else
		{					
			helpinfo();
		}
	}

	char* dptr = strchr(cptr, '#');
	if (dptr)
	{
		debug.open ("gprocess.dbg"); 

		if (!debug)
		{
			cout << "\nError opening debug file\n";
			exit(0);
		}
		else
		{
			debug << "Debug information from gprocess\n\n";
			debug << boolalpha;
		}

		if (!debug_setup(cptr))
		{
			debug_control::DEBUG = debug_control::SUMMARY;
			debug << "gprocess: default debug options set" << endl;
		}
		
	}

	if (strchr(cptr, 'c'))
	{
		RETAIN_CLASS_IDENTIFIERS = true;
	}

	if (strchr(cptr, 'e'))
	{
		WRITE_EQUIVALENCE_CLASSES = true;
	}

	if (strchr(cptr, 'h'))
		helpinfo();

	if (strchr(cptr, 'u'))
	{
		UNORIENTED_SEARCH = true;
	}

	if (strchr(cptr, 'p'))
	{
		WRITE_ELEMENT_POINTERS = true;
	}

	if (strchr(cptr, 'X'))
	{
		DEVELOPMENT_MODE = true;
	}
}

void set_programme_long_option(char* cptr, string source)
{
if (debug_control::DEBUG >= debug_control::SUMMARY)
	debug << "set_programme_long_option: passed " << cptr << " from " << source << endl;

	char  loc_buf[40];
	char* c1;
	char* c2;

	c1 = cptr;

	/* take out any leading -- */
	while (*c1 == '-')
		c1++;
	
	c2 = loc_buf;
	while (isalpha(*c1) || isdigit(*c1) || *c1 == '-' || *c1 == '_')
		*c2++ = *c1++;
    *c2 = '\0';
	
	if (!strcmp(loc_buf,"filepath"))
	{
		if (*c1 == '=')
		{
			filepath = string(++c1);
		}
		else
		{
			cout << "\nYou must specify a directory path with the filepath option, e.g. filepath=<directory-path>" << endl;
			exit(0);
		}
if (debug_control::DEBUG >= debug_control::SUMMARY)
	debug << "set_programme_long_option: filepath read from " << source << ", filepath = " << filepath << endl;
	}
	else if (!strcmp(loc_buf,"hash-threshold"))
	{
		if (*c1 == '=')
		{
			get_number(hash_threshold,++c1);
		}
		else
		{
			cout << "\nYou must specify the hash threshold with the num-cpts option, e.g. hash-threshold=1000" << endl;
			exit(0);
		}
		
if (debug_control::DEBUG >= debug_control::SUMMARY)
	debug << "set_programme_long_option: num-cpts " << source << ", num-cpts = " << max_num_crossings << endl;
	}
	else if (!strcmp(loc_buf,"num-cpts"))
	{
		if (*c1 == '=')
		{
			get_number(num_components,++c1);
		}
		else
		{
			cout << "\nYou must specify the number of components with the num-cpts option, e.g. num-cpts=2" << endl;
			exit(0);
		}
		
if (debug_control::DEBUG >= debug_control::SUMMARY)
	debug << "set_programme_long_option: num-cpts " << source << ", num-cpts = " << max_num_crossings << endl;
	}
	else if (!strcmp(loc_buf,"max-crossings"))
	{
		if (*c1 == '=')
		{
			get_number(max_num_crossings,++c1);
		}
		else
		{
			cout << "\nYou must specify the maximum number of crossings with the max-crossings option, e.g. max-crossings=7" << endl;
			exit(0);
		}
		
if (debug_control::DEBUG >= debug_control::SUMMARY)
	debug << "set_programme_long_option: max-crossings " << source << ", max-crossings = " << max_num_crossings << endl;
	}
	else if (!strcmp(loc_buf,"shift-power"))
	{
		if (*c1 == '=')
		{
			get_number(shift_power,++c1);
		}
		else
		{
			cout << "\nYou must specify the shift power with the shift-power option, e.g. shift-power=20" << endl;
			exit(0);
		}
		
if (debug_control::DEBUG >= debug_control::SUMMARY)
	debug << "set_programme_long_option: shift-power " << source << ", shift-power = " << shift_power << endl;
	}
	else
	{
if (debug_control::DEBUG >= debug_control::SUMMARY)
	debug << "set_programme_long_option: invalid long option " << loc_buf << " read from " << source << endl;
	
		cout << "Invalid long option " << loc_buf << endl;
		exit(0);		
	}
}

void print_prog_params (ostream& os, string level, string prefix)
{
    os << prefix << "UNORIENTED_SEARCH = " << UNORIENTED_SEARCH << endl;
    os << prefix << "WRITE_ELEMENT_POINTERS = " << WRITE_ELEMENT_POINTERS << endl;
    os << prefix << "max_num_crossings = " << max_num_crossings << endl;
    os << prefix << "num_components = " << num_components << endl;
    os << prefix << "filepath = " << filepath << endl;
    os << prefix << "hash_threshold = " << hash_threshold << endl;
    os << prefix << "shift_power = " << shift_power << endl;
}
